'''
'''
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os
import subprocess
import sys
from ports import get_port
import re

Test.Summary = '''
Test the reported type of HTTP transactions and tunnels
'''

# Define default ATS. Disable the cache to simplify the test.
ts = Test.MakeATSProcess("ts", enable_cache=False, enable_tls=True)
ts.addSSLfile("../tls/ssl/server.pem")
ts.addSSLfile("../tls/ssl/server.key")

server = Test.MakeOriginServer("server", ssl=True)
server2 = Test.MakeOriginServer("server2")

Test.testName = ""
request_tunnel_header = {"headers": "GET / HTTP/1.1\r\nHost: tunnel-test\r\n\r\n", "timestamp": "1469733493.993", "body": ""}
# expected response from the origin server
response_tunnel_header = {
    "headers": "HTTP/1.1 200 OK\r\nConnection: close\r\nContent-Length:0\r\n\r\n",
    "timestamp": "1469733493.993",
    "body": ""
}

Test.PrepareTestPlugin(os.path.join(Test.Variables.AtsTestPluginsDir, 'tunnel_transform.so'), ts)

# add response to the server dictionary
server.addResponse("sessionfile.log", request_tunnel_header, response_tunnel_header)
ts.Disk.records_config.update(
    {
        'proxy.config.diags.debug.enabled': 0,
        'proxy.config.diags.debug.tags': 'http|test',
        'proxy.config.ssl.server.cert.path': '{0}'.format(ts.Variables.SSLDir),
        'proxy.config.ssl.server.private_key.path': '{0}'.format(ts.Variables.SSLDir),
        'proxy.config.ssl.client.verify.server.policy': 'PERMISSIVE',
        'proxy.config.http.connect_ports': '{0}'.format(server.Variables.SSL_Port)
    })

ts.Disk.ssl_multicert_config.AddLine('dest_ip=* ssl_cert_name=server.pem ssl_key_name=server.key')

ts.Disk.sni_yaml.AddLines([
    'sni:',
    '- fqdn: tunnel-test',
    "  tunnel_route: localhost:{0}".format(server.Variables.SSL_Port),
])

# Set up simple forwarding proxy to keep track of TLS bytes for both
# directions
tr = Test.AddTestRun("Run dumb proxy and send tunnel request.")
tr.Setup.CopyAs('dumb_proxy.py', tr.RunDirectory)
dumb_proxy = tr.Processes.Process(f'dumb-proxy')
proxy_port = get_port(dumb_proxy, "listening_port")
dumb_proxy.Command = f'{sys.executable} dumb_proxy.py --listening_port {proxy_port} --forwarding_port {ts.Variables.ssl_port}'
dumb_proxy.StartBefore(Test.Processes.ts)
dumb_proxy.Ready = When.PortOpenv4(proxy_port)
dumb_proxy.ReturnCode = 0
# Record the log file path for later verification.
proxy_output = dumb_proxy.Streams.stdout.AbsPath

# Add connection close to ensure that the client connection closes promptly after completing the transaction
cmd_tunnel = 'curl -k --http1.1 -H "Connection: close" -vs --resolve "tunnel-test:{0}:127.0.0.1"  https://tunnel-test:{0}/'.format(
    proxy_port)

# Send the tunnel request
tr.Processes.Default.Env = ts.Env
tr.Processes.Default.Command = cmd_tunnel
tr.Processes.Default.ReturnCode = 0
tr.TimeOut = 5
tr.Processes.Default.StartBefore(server, ready=When.PortOpen(server.Variables.SSL_Port))
tr.Processes.Default.StartBefore(Test.Processes.ts)
tr.Processes.Default.StartBefore(dumb_proxy)
tr.StillRunningAfter = ts
tr.StillRunningAfter = server

# Signal that all the curl processes have completed
tr = Test.AddTestRun("Curl Done")
tr.DelayStart = 2  # Delaying a couple seconds to make sure the global continuation's lock contention resolves.
tr.Processes.Default.Command = "traffic_ctl plugin msg done done"
tr.Processes.Default.ReturnCode = 0
tr.Processes.Default.Env = ts.Env
tr.StillRunningAfter = ts
tr.StillRunningAfter = server

# Parking this as a ready tester on a meaningless process
# To stall the test runs that check for the stats until the
# stats have propagated and are ready to read.


def make_done_stat_ready(tsenv):

    def done_stat_ready(process, hasRunFor, **kw):
        retval = subprocess.run(
            "traffic_ctl metric get tunnel_transform.test.done",
            shell=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            env=tsenv)
        return b'1' in retval.stdout

    return done_stat_ready


# number of sessions/transactions opened and closed are equal
tr = Test.AddTestRun("Check type errors")
server2.StartupTimeout = 60
# Again, here the important thing is the ready function not the server2 process
tr.Processes.Default.StartBefore(server2, ready=make_done_stat_ready(ts.Env))
tr.Processes.Default.Command = 'traffic_ctl metric get tunnel_transform.error'
tr.Processes.Default.ReturnCode = 0
tr.Processes.Default.Env = ts.Env
tr.Processes.Default.Streams.All = Testers.ContainsExpression(
    'tunnel_transform.error 0', 'incorrect statistic return, or possible error.')
tr.StillRunningAfter = ts
tr.StillRunningAfter = server


def get_expected_bytes(path_to_proxy_output, key):
    # Construct the regex pattern.
    pattern = re.compile(rf'{re.escape(key)}:\s+(\d+)')
    with open(path_to_proxy_output, 'r') as file:
        log_content = file.read()
    bytes_transferred = 0
    match = pattern.search(log_content)
    if match:
        bytes_transferred = int(match.group(1))
    return bytes_transferred


def check_byte_count(plugin_metric_path, proxy_output_path, key):
    expected_bytes = get_expected_bytes(proxy_output_path, key)
    f = open(plugin_metric_path, 'r')
    content = f.read()
    values = content.split()
    f.close()
    if len(values) == 2:
        val = int(values[1])
        return val == expected_bytes, "Check byte count", "Byte count does not match the expected value"
    else:
        return False, "Check byte count", "Unexpected metrics output format"


tr = Test.AddTestRun("Fetch bytes sent")
tr.Processes.Default.Command = "traffic_ctl metric get tunnel_transform.ua.bytes_sent"
tr.Processes.Default.ReturnCode = 0
tr.Processes.Default.Env = ts.Env
tr.StillRunningAfter = ts
tr.StillRunningAfter = server

path1 = tr.Processes.Default.Streams.stdout.AbsPath

tr2 = Test.AddTestRun("Check the input bytes sent and fetch outptut bytes sent")
tr2.Processes.Default.Command = 'traffic_ctl metric get tunnel_transform.os.bytes_sent'
tr2.Processes.Default.ReturnCode = 0
tr2.Processes.Default.Env = ts.Env
tr2.StillRunningAfter = ts
tr2.StillRunningAfter = server
tr2.Processes.Default.Streams.stdout = Testers.Lambda(
    lambda info, tester: check_byte_count(path1, proxy_output, 'client-to-server'))

path2 = tr2.Processes.Default.Streams.stdout.AbsPath

tr = Test.AddTestRun("Check that output bytes sent")
tr.Processes.Default.Command = 'echo foo'
tr.Processes.Default.ReturnCode = 0
tr.Processes.Default.Env = ts.Env
tr.StillRunningAfter = ts
tr.StillRunningAfter = server
tr.Processes.Default.Streams.stdout = Testers.Lambda(lambda info, tester: check_byte_count(path2, proxy_output, 'server-to-client'))



'''
Verify ATS slice plugin config: @pparam=--strip-range-for-head
'''
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

Test.Summary = '''
Verify ATS slice plugin config: @pparam=--strip-range-for-head
'''
Test.SkipUnless(
    Condition.PluginExists('slice.so'),
    Condition.PluginExists('cache_range_requests.so'),
)


class SliceStripRangeForHeadRequestTest:
    replay_file = "replay/slice_range.replay.yaml"

    def __init__(self):
        """Initialize the Test processes for the test runs."""
        self._server = Test.MakeVerifierServerProcess("server", SliceStripRangeForHeadRequestTest.replay_file)
        self._configure_trafficserver()

    def _configure_trafficserver(self):
        """Configure Traffic Server."""
        self._ts = Test.MakeATSProcess("ts", enable_cache=False)

        self._ts.Disk.remap_config.AddLines(
            [
                f"map /no/range http://127.0.0.1:{self._server.Variables.http_port} \
                @plugin=slice.so @pparam=--blockbytes-test=10 @pparam=--strip-range-for-head \
                @plugin=cache_range_requests.so",
                f"map /with/range http://127.0.0.1:{self._server.Variables.http_port} \
                @plugin=slice.so @pparam=--blockbytes-test=10 \
                @plugin=cache_range_requests.so",
            ])

        self._ts.Disk.records_config.update(
            {
                'proxy.config.diags.debug.enabled': 1,
                'proxy.config.diags.debug.tags': 'http|slice|cache_range_requests',
            })

    def _test_head_request_range_header(self):
        tr = Test.AddTestRun()

        tr.AddVerifierClientProcess("client", SliceStripRangeForHeadRequestTest.replay_file, http_ports=[self._ts.Variables.port])

        tr.Processes.Default.StartBefore(self._server)
        tr.Processes.Default.StartBefore(self._ts)

    def run(self):
        self._test_head_request_range_header()


SliceStripRangeForHeadRequestTest().run()



'''
Test ATS configured to not respond with chunked content.
'''
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

Test.Summary = 'Test ATS configured to not respond with chunked content.'


class ChunkedEncodingDisabled:
    chunkedReplayFile = "replays/chunked.replay.yaml"

    def __init__(self):
        self.setupOriginServer()
        self.setupTS()

    def setupOriginServer(self):
        self.server = Test.MakeVerifierServerProcess("server", self.chunkedReplayFile)

    def setupTS(self):
        self.ts = Test.MakeATSProcess("ts", enable_tls=True, enable_cache=False)
        self.ts.addDefaultSSLFiles()
        self.ts.Disk.records_config.update(
            {
                "proxy.config.diags.debug.enabled": 1,
                "proxy.config.diags.debug.tags": "http",
                "proxy.config.ssl.server.cert.path": f'{self.ts.Variables.SSLDir}',
                "proxy.config.ssl.server.private_key.path": f'{self.ts.Variables.SSLDir}',
                "proxy.config.ssl.client.verify.server.policy": 'PERMISSIVE',

                # Never respond with chunked encoding.
                "proxy.config.http.chunking_enabled": 0,
            })
        self.ts.Disk.ssl_multicert_config.AddLine('dest_ip=* ssl_cert_name=server.pem ssl_key_name=server.key')
        self.ts.Disk.remap_config.AddLines(
            [
                f"map /for/http http://127.0.0.1:{self.server.Variables.http_port}/",
                f"map /for/tls https://127.0.0.1:{self.server.Variables.https_port}/",
            ])

    def runChunkedTraffic(self):
        tr = Test.AddTestRun()
        tr.AddVerifierClientProcess(
            "client", self.chunkedReplayFile, http_ports=[self.ts.Variables.port], https_ports=[self.ts.Variables.ssl_port])
        tr.Processes.Default.Streams.stdout += "gold/verifier_client_chunked.gold"

        tr.Processes.Default.StartBefore(self.server)
        tr.Processes.Default.StartBefore(self.ts)
        tr.StillRunningAfter = self.server
        tr.StillRunningAfter = self.ts

    def run(self):
        self.runChunkedTraffic()


ChunkedEncodingDisabled().run()



'''
'''
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os
import sys

Test.Summary = '''
Test new "all headers" log fields
'''

# Define ATS.
#
ts = Test.MakeATSProcess("ts")

# Define MicroServer.
#
server = Test.MakeOriginServer("server")

request_header = {"headers": "GET / HTTP/1.1\r\nHost: does.not.matter\r\n\r\n", "timestamp": "1469733493.993", "body": ""}
response_header = {
    "headers": "HTTP/1.1 200 OK\r\nConnection: close\r\nCache-control: max-age=85000\r\n\r\n",
    "timestamp": "1469733493.993",
    "body": "xxx"
}
server.addResponse("sessionlog.json", request_header, response_header)

ts.Disk.records_config.update({
    'proxy.config.diags.debug.enabled': 0,
    'proxy.config.diags.debug.tags': 'http|dns',
})

ts.Disk.remap_config.AddLine('map http://127.0.0.1:{0} http://127.0.0.1:{1}'.format(ts.Variables.port, server.Variables.Port))

# Mix in a numeric log field.  Hopefull this will detect any binary alignment problems.
#
ts.Disk.logging_yaml.AddLines(
    '''
logging:
  formats:
    - name: custom
      format: "%<cqah> ___FS___ %<pssc> ___FS___ %<psah> ___FS___ %<ssah> ___FS___ %<pqah> ___FS___ %<cssah> ___FS___ END_TXN"
  logs:
    - filename: test_all_headers
      format: custom
'''.split("\n"))

# Configure comparison of "sanitized" log file with gold file at end of test.
#
sanitized_log_path = os.path.join(ts.Variables.LOGDIR, 'test_all_headers.log.san')
Test.Disk.File(sanitized_log_path, exists=True, content='gold/test_all_headers.gold')


def reallyLong():
    value = 'abcdefghijklmnop'
    value = value + value
    value = value + value
    value = value + value
    retval = ""
    for i in range(3):
        retval += ' -H "x-header{}: {}"'.format(i, value)
    return retval


tr = Test.AddTestRun()
tr.Processes.Default.StartBefore(server)
tr.Processes.Default.StartBefore(Test.Processes.ts)
tr.Processes.Default.Command = (
    'curl "http://127.0.0.1:{0}" --user-agent "007" --verbose '.format(ts.Variables.port) + reallyLong())
tr.Processes.Default.ReturnCode = 0

# Repeat same curl, will be answered from the ATS cache.
#
tr = Test.AddTestRun()
tr.Processes.Default.Command = (
    'curl "http://127.0.0.1:{0}" --user-agent "007" --verbose '.format(ts.Variables.port) + reallyLong())
tr.Processes.Default.ReturnCode = 0

# Delay to allow TS to flush report to disk, then "sanitize" generated log.
#
tr = Test.AddTestRun()

log_path = os.path.join(ts.Variables.LOGDIR, 'test_all_headers.log')
sanitizer_python_script = os.path.join(Test.TestDirectory, 'all_headers_sanitizer.py')
sanitizer_shell_script = os.path.join(Test.TestDirectory, 'all_headers_sanitizer.sh')

tr.Processes.Default.Command = \
    (f'{sys.executable} {sanitizer_python_script} {log_path} {server.Variables.Port} | '
     f'sh {sanitizer_shell_script} > {sanitized_log_path}')
tr.Processes.Default.ReturnCode = 0



#!/usr/bin/env python3

#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
'''
A simple tool to generate some raw http2 frames for 0-rtt testing.
'''

# http2 frame format:
# +-----------------------------------------------+
# |                 Length (24)                   |
# +---------------+---------------+---------------+
# |   Type (8)    |   Flags (8)   |
# +-+-------------+---------------+-------------------------------+
# |R|                 Stream Identifier (31)                      |
# +=+=============================================================+
# |                   Frame Payload (0...)                      ...
# +---------------------------------------------------------------+

import hpack
import os
import sys

H2_PREFACE = bytes.fromhex('505249202a20485454502f322e300d0a0d0a534d0d0a0d0a')

RESERVED_BIT_MASK = 0x7FFFFFFF

TYPE_HEADERS_FRAME = 0x01
TYPE_SETTINGS_FRAME = 0x04
TYPE_WINDOW_UPDATE_FRAME = 0x08

SETTINGS_HEADER_TABLE_SIZE = 0x01
SETTINGS_ENABLE_PUSH = 0x02
SETTINGS_MAX_CONCURRENT_STREAMS = 0x03
SETTINGS_INITIAL_WINDOW_SIZE = 0x04
SETTINGS_MAX_FRAME_SIZE = 0x05
SETTINGS_MAX_HEADER_LIST_SIZE = 0x06

HEADERS_FLAG_END_STREAM = 0x01
HEADERS_FLAG_END_HEADERS = 0x04
HEADERS_FLAG_END_PADDED = 0x08
HEADERS_FLAG_END_PRIORITY = 0x20

CURRENT_SCRIPT_PATH = os.path.dirname(os.path.abspath(__file__))


def encode_payload(data):
    encoder = hpack.Encoder()
    data_encoded = encoder.encode(data)
    return data_encoded


def make_frame(frame_length, frame_type, frame_flags, frame_stream_id, frame_payload):
    frame_length = bytes.fromhex('{0:06x}'.format(frame_length))
    frame_type = bytes.fromhex('{0:02x}'.format(frame_type))
    frame_flags = bytes.fromhex('{0:02x}'.format(frame_flags))
    frame_stream_id = bytes.fromhex('{0:08x}'.format(RESERVED_BIT_MASK & frame_stream_id))

    frame = frame_length + frame_type + frame_flags + frame_stream_id

    if frame_payload is not None:
        frame += frame_payload

    return frame


def make_settins_frame(ack=False, empty=False):
    payload = ''
    if not ack and not empty:
        payload += '{0:04x}{1:08x}'.format(SETTINGS_ENABLE_PUSH, 0)
        payload += '{0:04x}{1:08x}'.format(SETTINGS_MAX_CONCURRENT_STREAMS, 100)
        payload += '{0:04x}{1:08x}'.format(SETTINGS_INITIAL_WINDOW_SIZE, 1073741824)
    payload = bytes.fromhex(payload)

    frame = make_frame(
        frame_length=len(payload),
        frame_type=TYPE_SETTINGS_FRAME,
        frame_flags=1 if ack else 0,
        frame_stream_id=0,
        frame_payload=payload)

    return frame


def make_window_update_frame():
    payload = '{0:08x}'.format(RESERVED_BIT_MASK & 1073676289)
    payload = bytes.fromhex(payload)

    frame = make_frame(
        frame_length=len(payload), frame_type=TYPE_WINDOW_UPDATE_FRAME, frame_flags=0, frame_stream_id=0, frame_payload=payload)
    return frame


def make_headers_frame(method, path='', stream_id=0x01):
    headers = []
    if method == 'get':
        headers.append((':method', 'GET'))
        if path != '':
            headers.append((':path', path))
        else:
            headers.append((':path', '/early_get'))
    elif method == 'post':
        headers.append((':method', 'POST'))
        if path != '':
            headers.append((':path', path))
        else:
            headers.append((':path', '/early_post'))

    headers.extend([(':scheme', 'https'), (':authority', '127.0.0.1'), ('host', '127.0.0.1'), ('accept', '*/*')])

    headers_encoded = encode_payload(headers)

    frame = make_frame(
        frame_length=len(headers_encoded),
        frame_type=TYPE_HEADERS_FRAME,
        frame_flags=HEADERS_FLAG_END_STREAM | HEADERS_FLAG_END_HEADERS,
        frame_stream_id=stream_id,
        frame_payload=headers_encoded)

    return frame


def make_h2_req(test):
    h2_req = H2_PREFACE
    if test == 'get' or test == 'post':
        frames = [make_settins_frame(ack=True), make_headers_frame(test)]
        for frame in frames:
            h2_req += frame
    elif test == 'multi1':
        frames = [
            make_settins_frame(ack=True),
            make_headers_frame('get', '/early_multi_1', 1),
            make_headers_frame('get', '/early_multi_2', 3),
            make_headers_frame('get', '/early_multi_3', 5)
        ]
        for frame in frames:
            h2_req += frame
    elif test == 'multi2':
        frames = [
            make_settins_frame(ack=True),
            make_headers_frame('get', '/early_multi_1', 1),
            make_headers_frame('post', stream_id=3),
            make_headers_frame('get', '/early_multi_3', 5)
        ]
        for frame in frames:
            h2_req += frame
    else:
        pass
    return h2_req


def write_to_file(data, file_name):
    with open(file_name, 'wb') as out_file:
        out_file.write(data)
    return


def main():
    test = sys.argv[1]
    write_to_file(make_h2_req(test), os.path.join(CURRENT_SCRIPT_PATH, 'early_h2_{0}.txt'.format(test)))
    exit(0)


if __name__ == '__main__':
    main()



'''
Test certifier plugin behaviors
'''
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the #  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import hashlib
import os
import re

Test.Summary = '''
Test certifier plugin behaviors
'''

Test.SkipUnless(Condition.PluginExists('certifier.so'))


class DynamicCertTest:
    httpsReplayFile = "replays/https.replay.yaml"
    certPathSrc = os.path.join(Test.TestDirectory, "certs")
    host = "www.tls.com"
    certPathDest = ""

    def __init__(self):
        self.setupOriginServer()
        self.setupTS()

    def setupOriginServer(self):
        self.server = Test.MakeVerifierServerProcess("verifier-server1", self.httpsReplayFile)

    def setupTS(self):
        self.ts = Test.MakeATSProcess("ts1", enable_tls=True)
        self.ts.addDefaultSSLFiles()
        # copy over the cert store in which the certs will be generated/stored
        self.certPathDest = os.path.join(self.ts.Variables.CONFIGDIR, "certifier-certs")
        Setup.Copy(self.certPathSrc, self.certPathDest)
        Setup.MakeDir(os.path.join(self.certPathDest, 'store'))
        self.ts.Disk.records_config.update(
            {
                "proxy.config.diags.debug.enabled": 1,
                "proxy.config.diags.debug.tags": "http|certifier|ssl",
                "proxy.config.ssl.server.cert.path": f'{self.ts.Variables.SSLDir}',
                "proxy.config.ssl.server.private_key.path": f'{self.ts.Variables.SSLDir}',
            })
        self.ts.Disk.ssl_multicert_config.AddLine('dest_ip=* ssl_cert_name=server.pem ssl_key_name=server.key')
        self.ts.Disk.remap_config.AddLine(f"map / http://127.0.0.1:{self.server.Variables.http_port}/",)
        self.ts.Disk.plugin_config.AddLine(
            f'certifier.so -s {os.path.join(self.certPathDest, "store")} -m 1000 -c {os.path.join(self.certPathDest, "ca.cert")} -k {os.path.join(self.certPathDest, "ca.key")} -r {os.path.join(self.certPathDest, "ca-serial.txt")}'
        )
        # Verify logs for dynamic generation of certs
        self.ts.Disk.traffic_out.Content += Testers.ContainsExpression(
            "creating shadow certs", "Verify the certifier plugin generates the certificate dynamically.")

    def runHTTPSTraffic(self):
        tr = Test.AddTestRun("Test dynamic generation of certs")
        tr.AddVerifierClientProcess(
            "client1",
            self.httpsReplayFile,
            http_ports=[self.ts.Variables.port],
            https_ports=[self.ts.Variables.ssl_port],
            other_args='--thread-limit 1')
        tr.Processes.Default.StartBefore(self.server)
        tr.Processes.Default.StartBefore(self.ts)
        tr.StillRunningAfter = self.server
        tr.StillRunningAfter = self.ts

    def verifyCert(self, certPath):
        tr = Test.AddTestRun("Verify the content of the generated cert")
        tr.Processes.Default.Command = f'openssl x509 -in {certPath} -text -noout'
        tr.Processes.Default.ReturnCode = 0
        # Verify certificate content
        tr.Processes.Default.Streams.All += Testers.ContainsExpression(
            "Subject: CN ?= ?www.tls.com", "Subject should match the host in the request")
        tr.Processes.Default.Streams.All += Testers.ContainsExpression(
            r"X509v3 extensions:\n.*X509v3 Subject Alternative Name:.*\n.*DNS:www.tls.com",
            "Should contain the SAN extension",
            reflags=re.MULTILINE)

    def verifyCertNotExist(self, certPath):
        tr = Test.AddTestRun("Verify the cert doesn't exist in the store")
        tr.Processes.Default.Command = "echo verify"
        certFile = tr.Disk.File(certPath, exists=False)

    def run(self):
        # the certifier plugin generates the cert and store it in a directory
        # named with the first three character of the md5 hash of the hostname
        genCertPath = os.path.join(
            self.certPathDest, 'store', str(hashlib.md5(self.host.encode('utf-8')).hexdigest()[:3]), self.host + ".crt")
        self.verifyCertNotExist(genCertPath)
        self.runHTTPSTraffic()
        self.verifyCert(genCertPath)


DynamicCertTest().run()


class ReuseExistingCertTest:
    httpsReplayFile = "replays/https-two-sessions.replay.yaml"
    certPathSrc = os.path.join(Test.TestDirectory, "certs")
    host = "www.tls.com"
    certPathDest = ""

    def __init__(self):
        self.setupOriginServer()
        self.setupTS()

    def setupOriginServer(self):
        self.server = Test.MakeVerifierServerProcess("verifier-server2", self.httpsReplayFile)

    def setupTS(self):
        self.ts = Test.MakeATSProcess("ts2", enable_tls=True)
        self.ts.addDefaultSSLFiles()
        # copy over the cert store in which the certs will be generated/stored
        self.certPathDest = os.path.join(self.ts.Variables.CONFIGDIR, "certifier-certs")
        Setup.Copy(self.certPathSrc, self.certPathDest)
        Setup.MakeDir(os.path.join(self.certPathDest, 'store'))
        self.ts.Disk.records_config.update(
            {
                "proxy.config.diags.debug.enabled": 1,
                "proxy.config.diags.debug.tags": "http|certifier|ssl",
                "proxy.config.ssl.server.cert.path": f'{self.ts.Variables.SSLDir}',
                "proxy.config.ssl.server.private_key.path": f'{self.ts.Variables.SSLDir}',
            })
        self.ts.Disk.ssl_multicert_config.AddLine('dest_ip=* ssl_cert_name=server.pem ssl_key_name=server.key')
        self.ts.Disk.remap_config.AddLine(f"map / http://127.0.0.1:{self.server.Variables.http_port}/",)
        self.ts.Disk.plugin_config.AddLine(
            f'certifier.so -s {os.path.join(self.certPathDest, "store")} -m 1000 -c {os.path.join(self.certPathDest, "ca.cert")} -k {os.path.join(self.certPathDest, "ca.key")} -r {os.path.join(self.certPathDest, "ca-serial.txt")}'
        )
        # Verify logs for reusing existing cert
        self.ts.Disk.traffic_out.Content += Testers.ContainsExpression(
            "reusing existing cert and context for www.tls.com", "Should reuse the existing certificate")

    def runHTTPSTraffic(self):
        tr = Test.AddTestRun("Test dynamic generation of certs")
        tr.AddVerifierClientProcess(
            "client2",
            self.httpsReplayFile,
            http_ports=[self.ts.Variables.port],
            https_ports=[self.ts.Variables.ssl_port],
            other_args='--thread-limit 1')
        tr.Processes.Default.StartBefore(self.server)
        tr.Processes.Default.StartBefore(self.ts)
        tr.StillRunningAfter = self.server
        tr.StillRunningAfter = self.ts

    def run(self):
        self.runHTTPSTraffic()


ReuseExistingCertTest().run()



'''
'''
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os

Test.Summary = '''
Test tls server prefetched OCSP responses
'''

Test.SkipUnless(
    Condition.HasCurlVersion("7.41.0"),  # curl --cert-status option has been introduced in version 7.41.0
)

# Define default ATS
ts = Test.MakeATSProcess("ts", enable_tls=True)
server = Test.MakeOriginServer("server")
request_header = {"headers": "GET / HTTP/1.1\r\nHost: example.com\r\n\r\n", "timestamp": "1469733493.993", "body": ""}
# desired response form the origin server
response_header = {"headers": "HTTP/1.1 200 OK\r\nConnection: close\r\n\r\n", "timestamp": "1469733493.993", "body": ""}
server.addResponse("sessionlog.json", request_header, response_header)

ts.addSSLfile("ssl/ca.ocsp.pem")
ts.addSSLfile("ssl/server.ocsp.pem")
ts.addSSLfile("ssl/server.ocsp.key")
ts.addSSLfile("ssl/ocsp_response.der")

ts.Disk.remap_config.AddLine(
    'map https://example.com:{0} http://127.0.0.1:{1}'.format(ts.Variables.ssl_port, server.Variables.Port))

ts.Disk.ssl_multicert_config.AddLine(
    'dest_ip=* ssl_cert_name=server.ocsp.pem ssl_key_name=server.ocsp.key ssl_ocsp_name=ocsp_response.der')

# Case 1, global config policy=permissive properties=signature
#         override for foo.com policy=enforced properties=all
ts.Disk.records_config.update(
    {
        'proxy.config.ssl.server.cert.path': '{0}'.format(ts.Variables.SSLDir),
        'proxy.config.ssl.server.private_key.path': '{0}'.format(ts.Variables.SSLDir),
        'proxy.config.ssl.server.cert_chain.filename': 'ca.ocsp.pem',
        # enable prefetched OCSP responses
        'proxy.config.ssl.ocsp.response.path': '{0}'.format(ts.Variables.SSLDir),
        'proxy.config.ssl.ocsp.enabled': 1,
        'proxy.config.exec_thread.autoconfig.scale': 1.0,
        'proxy.config.diags.debug.enabled': 1,
        'proxy.config.diags.debug.tags': 'ssl_ocsp'
    })

tr = Test.AddTestRun("Check OCSP response using curl")
tr.Processes.Default.StartBefore(server)
tr.Processes.Default.StartBefore(Test.Processes.ts)
tr.StillRunningAfter = server
tr.StillRunningAfter = ts
tr.Processes.Default.Command = "curl -v --cacert {0} --cert-status -H \"host:example.com\" https://127.0.0.1:{1}".format(
    os.path.join(ts.Variables.SSLDir, "ca.ocsp.pem"), ts.Variables.ssl_port)
tr.ReturnCode = 0



'''
Server receives POST, sent by client over HTTP/2, waits 2 minutes, then sends response with 200 KBytes of data
'''
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

Test.Summary = '''
Server receives POST, sent by client over HTTP/2, waits 2 minutes, then sends response with 200 KBytes of data
'''

# Because of the 2 minute delay, we don't want to run this test in CI checks.  Comment out this line to run it.
Test.SkipIf(Condition.true("Test takes too long to run it in CI."))

Test.SkipUnless(Condition.HasCurlFeature('http2'))

ts = Test.MakeATSProcess("ts", enable_tls=True, enable_cache=False)

ts.addDefaultSSLFiles()

ts.Disk.records_config.update(
    {
        'proxy.config.diags.debug.enabled': 1,
        'proxy.config.diags.debug.tags': 'http',
        'proxy.config.proxy_name': 'Poxy_Proxy',  # This will be the server name.
        'proxy.config.ssl.server.cert.path': '{0}'.format(ts.Variables.SSLDir),
        'proxy.config.ssl.server.private_key.path': '{0}'.format(ts.Variables.SSLDir),
        'proxy.config.http.transaction_no_activity_timeout_out': 150,
        'proxy.config.http2.no_activity_timeout_in': 150,
    })

ts.Disk.ssl_multicert_config.AddLine('dest_ip=* ssl_cert_name=server.pem ssl_key_name=server.key')

Test.GetTcpPort("server_port")

ts.Disk.remap_config.AddLine('map https://localhost http://localhost:{}'.format(Test.Variables.server_port))

server = Test.Processes.Process("server", "bash -c '" + Test.TestDirectory + "/server.sh {}'".format(Test.Variables.server_port))

tr = Test.AddTestRun()
tr.Processes.Default.Command = (
    'curl --request POST --verbose --ipv4 --http2 --insecure --header "Content-Length: 0"' +
    " --header 'Host: localhost' https://localhost:{}/xyz >curl.log 2>curl.err".format(ts.Variables.ssl_port))
tr.Processes.Default.ReturnCode = 0
tr.Processes.Default.StartBefore(server)
tr.Processes.Default.StartBefore(ts)

# Make sure the curl command received 200 KB of data.
#
tr = Test.AddTestRun()
tr.Processes.Default.Command = "bash -c '" + Test.TestDirectory + "/check.sh'"
tr.Processes.Default.ReturnCode = 0



'''
Verify the behavior of proxy.config.http.per_server.connection.max.
'''
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

Test.Summary = __doc__


class PerServerConnectionMaxTest:
    """Define an object to test our max origin connection behavior."""

    _replay_file: str = 'slow_servers.replay.yaml'
    _origin_max_connections: int = 3

    def __init__(self) -> None:
        """Configure the test processes in preparation for the TestRun."""
        self._configure_dns()
        self._configure_server()
        self._configure_trafficserver()

    def _configure_dns(self) -> None:
        """Configure a nameserver for the test."""
        self._dns = Test.MakeDNServer("dns1", default='127.0.0.1')

    def _configure_server(self) -> None:
        """Configure the server to be used in the test."""
        self._server = Test.MakeVerifierServerProcess('server1', self._replay_file)

    def _configure_trafficserver(self) -> None:
        """Configure Traffic Server to be used in the test."""
        self._ts = Test.MakeATSProcess("ts1")
        self._ts.Disk.remap_config.AddLine(f'map / http://127.0.0.1:{self._server.Variables.http_port}')
        self._ts.Disk.records_config.update(
            {
                'proxy.config.dns.nameservers': f"127.0.0.1:{self._dns.Variables.Port}",
                'proxy.config.dns.resolv_conf': 'NULL',
                'proxy.config.diags.debug.enabled': 1,
                'proxy.config.diags.debug.tags': 'http',
                'proxy.config.http.per_server.connection.max': self._origin_max_connections,
            })
        self._ts.Disk.diags_log.Content += Testers.ContainsExpression(
            f'WARNING:.*too many connections:.*limit={self._origin_max_connections}',
            'Verify the user is warned about the connection limit being hit.')

    def run(self) -> None:
        """Configure the TestRun."""
        tr = Test.AddTestRun('Verify we enforce proxy.config.http.per_server.connection.max')
        tr.Processes.Default.StartBefore(self._dns)
        tr.Processes.Default.StartBefore(self._server)
        tr.Processes.Default.StartBefore(self._ts)

        tr.AddVerifierClientProcess('client', self._replay_file, http_ports=[self._ts.Variables.port])


class ConnectMethodTest:
    """Test our max origin connection behavior with CONNECT traffic."""

    _client_counter: int = 0
    _origin_max_connections: int = 3

    def __init__(self) -> None:
        """Configure the server processes in preparation for the TestRun."""
        self._configure_dns()
        self._configure_origin_server()
        self._configure_trafficserver()

    def _configure_dns(self) -> None:
        """Configure a nameserver for the test."""
        self._dns = Test.MakeDNServer("dns2", default='127.0.0.1')

    def _configure_origin_server(self) -> None:
        """Configure the httpbin origin server."""
        self._server = Test.MakeHttpBinServer("server2")

    def _configure_trafficserver(self) -> None:
        self._ts = Test.MakeATSProcess("ts2")

        self._ts.Disk.records_config.update(
            {
                'proxy.config.dns.nameservers': f"127.0.0.1:{self._dns.Variables.Port}",
                'proxy.config.dns.resolv_conf': 'NULL',
                'proxy.config.diags.debug.enabled': 1,
                'proxy.config.diags.debug.tags': 'http|dns|hostdb',
                'proxy.config.http.server_ports': f"{self._ts.Variables.port}",
                'proxy.config.http.connect_ports': f"{self._server.Variables.Port}",
                'proxy.config.http.per_server.connection.max': self._origin_max_connections,
            })

        self._ts.Disk.remap_config.AddLines([
            f"map http://foo.com/ http://www.this.origin.com:{self._server.Variables.Port}/",
        ])

    def _configure_client_with_slow_response(self, tr) -> 'Test.Process':
        """Configure a client to perform a CONNECT request with a slow response from the server."""
        p = tr.Processes.Process(f'slow_client_{self._client_counter}')
        self._client_counter += 1
        p.Command = f"curl -v --fail -s -p -x 127.0.0.1:{self._ts.Variables.port} 'http://foo.com/delay/2'"
        return p

    def run(self) -> None:
        """Verify per_server.connection.max with CONNECT traffic."""
        tr = Test.AddTestRun()
        tr.Processes.Default.StartBefore(self._dns)
        tr.Processes.Default.StartBefore(self._server)
        tr.Processes.Default.StartBefore(self._ts)

        slow0 = self._configure_client_with_slow_response(tr)
        slow1 = self._configure_client_with_slow_response(tr)
        slow2 = self._configure_client_with_slow_response(tr)

        tr.Processes.Default.StartBefore(slow0)
        tr.Processes.Default.StartBefore(slow1)
        tr.Processes.Default.StartBefore(slow2)

        # With those three slow transactions going on in the background, do a
        # couple quick transactions and make sure they both reply with a 503
        # response.
        tr.Processes.Default.Command = (
            f"sleep 1; curl -v --fail -s -p -x 127.0.0.1:{self._ts.Variables.port} 'http://foo.com/get'"
            f"--next -v --fail -s -p -x 127.0.0.1:{self._ts.Variables.port} 'http://foo.com/get'")
        # Curl will have a 22 exit code if it receives a 5XX response (and we
        # expect a 503).
        tr.Processes.Default.ReturnCode = 22
        tr.Processes.Default.Streams.stderr = "gold/two_503_congested.gold"
        tr.Processes.Default.TimeOut = 3


PerServerConnectionMaxTest().run()
ConnectMethodTest().run()



'''
A simple forwarding proxy that forwards all traffic from one local port to
another, while keeping track of the number of bytes transferred in each
direction.
'''
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import socket
import threading
import argparse

LOCAL_HOST = '127.0.0.1'
TIMEOUT = 0.5
# Create a thread-local data object to store the number of bytes transferred.
thread_local_data = threading.local()


def parse_args() -> argparse.Namespace:
    """Parse command line arguments."""
    parser = argparse.ArgumentParser(description=__doc__)

    parser.add_argument('--listening_port', type=int, help='Port where the proxy listens.')

    parser.add_argument('--forwarding_port', type=int, help='Server port to forward to.')

    return parser.parse_args()


def initialize_thread_local_data():
    thread_local_data.client_to_server_bytes = 0
    thread_local_data.server_to_client_bytes = 0


def forward(source, destination, is_client_to_server):
    """Forward traffic from source to destination.

    :param source: socket to read from.
    :param destination: socket to write to.
    :param is_client_to_server: True if forwarding from client to server.
    """
    # Initialize thread-local data.
    initialize_thread_local_data()

    while True:
        try:
            data = source.recv(4096)
            if not data:
                break
            destination.sendall(data)
        except Exception as e:
            # Catching all exceptions.
            break
        if is_client_to_server:
            thread_local_data.client_to_server_bytes += len(data)
        else:
            thread_local_data.server_to_client_bytes += len(data)
    # Forwarding done. Print the number of bytes transferred in the direction.
    if thread_local_data.client_to_server_bytes > 0:
        print(f"client-to-server: {thread_local_data.client_to_server_bytes}")
    elif thread_local_data.server_to_client_bytes > 0:
        print(f"server-to-client: {thread_local_data.server_to_client_bytes}")


def start_bidirectional_forwarding(client_socket, forwarding_port):
    """Start forwarding traffic between client and server.

    :param client_socket: socket connected to the client.
    :param forwarding_port: server port to forward to.
    """
    CLIENT_TO_SERVER = True
    SERVER_TO_CLIENT = False
    with client_socket, socket.socket(socket.AF_INET, socket.SOCK_STREAM) as server_socket:
        client_socket.settimeout(TIMEOUT)
        server_socket.settimeout(TIMEOUT)
        server_socket.connect((LOCAL_HOST, forwarding_port))
        # Spawn a thread to forward traffic from client to server.
        client_to_server = threading.Thread(target=forward, args=(client_socket, server_socket, CLIENT_TO_SERVER))
        client_to_server.start()

        # Forward traffic from server to client in the current thread.
        forward(server_socket, client_socket, SERVER_TO_CLIENT)
        client_to_server.join()


def main() -> int:
    """Run the proxy."""
    print(f"Starting proxy...")
    args = parse_args()
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as listen_socket:
        listen_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        listen_socket.bind((LOCAL_HOST, args.listening_port))
        listen_socket.listen()
        print(f"Proxy listening on {LOCAL_HOST}:{args.listening_port}")
        try:
            while True:
                client_sock, client_addr = listen_socket.accept()
                print(f"Accepted connection from {client_addr}")
                # Handle each client connection in a new thread.
                client_thread = threading.Thread(target=start_bidirectional_forwarding, args=(client_sock, args.forwarding_port))
                client_thread.start()
        except Exception:
            # Catching all exceptions.
            pass
        except KeyboardInterrupt:
            print("Caught KeyboardInterrupt, terminating the program")
            return 0


if __name__ == "__main__":
    main()



'''
Verify the behavior of proxy.config.net.per_client.connection.max.
'''
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from enum import Enum

Test.Summary = __doc__


class Protocol(Enum):
    HTTP = 1
    HTTPS = 2
    HTTP2 = 3

    @classmethod
    def to_str(cls, protocol: int) -> str:
        """Convert the protocol to a string.

        :param protocol: The protocol to convert.
        :return: The string representation of the protocol.
        """
        if protocol == cls.HTTP:
            return 'http'
        elif protocol == cls.HTTPS:
            return 'https'
        elif protocol == cls.HTTP2:
            return 'http2'
        else:
            raise ValueError(f'Unknown protocol: {protocol}')


class PerClientConnectionMaxTest:
    """Define an object to test our max client connection behavior."""

    _dns_counter: int = 0
    _server_counter: int = 0
    _ts_counter: int = 0
    _client_counter: int = 0
    _max_client_connections: int = 3
    _protocol_to_replay_file = {
        Protocol.HTTP: 'http_slow_origins.replay.yaml',
        Protocol.HTTPS: 'https_slow_origins.replay.yaml',
        Protocol.HTTP2: 'http2_slow_origins.replay.yaml',
    }

    def __init__(self, protocol: int) -> None:
        """Configure the test processes in preparation for the TestRun.

        :param protocol: The protocol to test.
        """
        self._protocol = protocol
        protocol_string = Protocol.to_str(protocol)
        self._replay_file = self._protocol_to_replay_file[protocol]
        tr = Test.AddTestRun(f'proxy.config.net.per_client.connection.max: {protocol_string}')
        self._configure_dns(tr)
        self._configure_server(tr)
        self._configure_trafficserver()
        self._configure_client(tr)
        self._verify_metrics()

    def _configure_dns(self, tr: 'TestRun') -> None:
        """Configure a nameserver for the test.

        :param tr: The TestRun to add the nameserver to.
        """
        name = f'dns{PerClientConnectionMaxTest._dns_counter}'
        self._dns = tr.MakeDNServer(name, default='127.0.0.1')
        PerClientConnectionMaxTest._dns_counter += 1

    def _configure_server(self, tr: 'TestRun') -> None:
        """Configure the server to be used in the test.

        :param tr: The TestRun to add the server to.
        """
        name = f'server{PerClientConnectionMaxTest._server_counter}'
        self._server = tr.AddVerifierServerProcess(name, self._replay_file)
        PerClientConnectionMaxTest._server_counter += 1
        self._server.Streams.All += Testers.ContainsExpression(
            "first-request", "Verify the first request should have been received.")
        self._server.Streams.All += Testers.ContainsExpression(
            "second-request", "Verify the second request should have been received.")
        self._server.Streams.All += Testers.ContainsExpression(
            "third-request", "Verify the third request should have been received.")
        self._server.Streams.All += Testers.ContainsExpression(
            "fifth-request", "Verify the fifth request should have been received.")

        # The fourth request should be blocked due to too many connections.
        self._server.Streams.All += Testers.ExcludesExpression(
            "fourth-request", "Verify the fourth request should not be received.")

    def _configure_trafficserver(self) -> None:
        """Configure Traffic Server to be used in the test."""
        # Associate ATS with the Test so that metrics can be verified.
        name = f'ts{PerClientConnectionMaxTest._ts_counter}'
        self._ts = Test.MakeATSProcess(name, enable_cache=False, enable_tls=True)
        PerClientConnectionMaxTest._ts_counter += 1
        self._ts.addDefaultSSLFiles()
        self._ts.Disk.ssl_multicert_config.AddLine('dest_ip=* ssl_cert_name=server.pem ssl_key_name=server.key')
        if self._protocol == Protocol.HTTP:
            server_port = self._server.Variables.http_port
            scheme = 'http'
        else:
            server_port = self._server.Variables.https_port
            scheme = 'https'
        self._ts.Disk.remap_config.AddLine(f'map / {scheme}://127.0.0.1:{server_port}')
        self._ts.Disk.records_config.update(
            {
                'proxy.config.ssl.server.cert.path': self._ts.Variables.SSLDir,
                'proxy.config.ssl.server.private_key.path': self._ts.Variables.SSLDir,
                'proxy.config.ssl.client.verify.server.policy': 'PERMISSIVE',
                'proxy.config.dns.nameservers': f"127.0.0.1:{self._dns.Variables.Port}",
                'proxy.config.dns.resolv_conf': 'NULL',
                'proxy.config.diags.debug.enabled': 1,
                'proxy.config.diags.debug.tags': 'socket|http|net_queue|iocore_net|conn_track',
                'proxy.config.net.per_client.max_connections_in': self._max_client_connections,
                # Disable keep-alive so we close the client connections when the
                # transactions are done. This allows us to verify cleanup is working
                # per the ConnectionTracker metrics.
                'proxy.config.http.keep_alive_enabled_in': 0,
            })
        self._ts.Disk.diags_log.Content += Testers.ContainsExpression(
            f'WARNING:.*too many connections:.*limit={self._max_client_connections}',
            'Verify the user is warned about the connection limit being hit.')

    def _configure_client(self, tr: 'TestRun') -> None:
        """Configure the TestRun.

        :param tr: The TestRun to add the client to.
        """
        name = f'client{PerClientConnectionMaxTest._client_counter}'
        p = tr.AddVerifierClientProcess(
            name, self._replay_file, http_ports=[self._ts.Variables.port], https_ports=[self._ts.Variables.ssl_port])
        PerClientConnectionMaxTest._client_counter += 1

        p.StartBefore(self._dns)
        p.StartBefore(self._server)
        p.StartBefore(self._ts)

        # Because the fourth connection will be aborted, the client will have a
        # non-zero return code.
        p.ReturnCode = 1
        p.Streams.All += Testers.ContainsExpression("first-request", "Verify the first request should have been received.")
        p.Streams.All += Testers.ContainsExpression("second-request", "Verify the second request should have been received.")
        p.Streams.All += Testers.ContainsExpression("third-request", "Verify the third request should have been received.")
        p.Streams.All += Testers.ContainsExpression("fifth-request", "Verify the fifth request should have been received.")
        if self._protocol == Protocol.HTTP:
            p.Streams.All += Testers.ContainsExpression(
                "The peer closed the connection while reading.",
                "A connection should be closed due to too many client connections.")
            p.Streams.All += Testers.ContainsExpression(
                "Failed HTTP/1 transaction with key: fourth-request", "The fourth request should fail.")
        else:
            p.Streams.All += Testers.ContainsExpression(
                "ECONNRESET: Connection reset by peer", "A connection should be closed due to too many client connections.")
            p.Streams.All += Testers.ExcludesExpression("fourth-request", "The fourth request should fail.")

    def _verify_metrics(self) -> None:
        """Verify the per client connection metrics."""
        tr = Test.AddTestRun("Verify the per client connection metrics.")
        tr.Processes.Default.Env = self._ts.Env
        tr.Processes.Default.Command = (
            'traffic_ctl metric get '
            'proxy.process.net.per_client.connections_throttled_in '
            'proxy.process.net.connection_tracker_table_size')
        tr.Processes.Default.ReturnCode = 0
        tr.Processes.Default.Streams.All += Testers.ContainsExpression(
            'proxy.process.net.per_client.connections_throttled_in 1', 'Verify the per client throttled metric is correct.')
        tr.Processes.Default.Streams.All += Testers.ContainsExpression(
            'proxy.process.net.connection_tracker_table_size 0', 'Verify the table was cleaned up correctly.')


PerClientConnectionMaxTest(Protocol.HTTP)
PerClientConnectionMaxTest(Protocol.HTTPS)
PerClientConnectionMaxTest(Protocol.HTTP2)



'''
Verify the behavior of the JA3 fingerprint plugin.
'''
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os
import re

Test.Summary = __doc__

# Skip if plugin not present. Plugin gets built only if openssl is used(not BoringSSL).
Test.SkipUnless(Condition.PluginExists('ja3_fingerprint.so'))


class JA3FingerprintTest:
    """Verify the behavior of the JA3 fingerprint plugin."""

    _dns_counter: int = 0
    _server_counter: int = 0
    _ts_counter: int = 0
    _client_counter: int = 0

    def __init__(self, test_remap: bool, modify_incoming: bool) -> None:
        """Configure the test processes in preparation for the TestRun.

        :param test_remap: Whether to configure the plugin as a remap plugin
        instead of as a global plugin.
        :param modify_incoming: Whether ja3_fingerprint should be configured to
        modify the client request rather than the proxy request.
        """
        if test_remap and modify_incoming:
            raise ValueError('modify-incoming is only allowed as a global plugin.')

        self._test_remap = test_remap
        if test_remap:
            self._replay_file = 'ja3_fingerprint_remap.replay.yaml'
        else:
            self._replay_file = 'ja3_fingerprint_global.replay.yaml'

        self._modify_incoming = modify_incoming

        tr = Test.AddTestRun('Testing ja3_fingerprint plugin.')
        self._configure_dns(tr)
        self._configure_server(tr)
        self._configure_trafficserver()
        self._configure_client(tr)
        self._await_ja3log()
        self._verify_internal_headers()

    def _configure_dns(self, tr: 'TestRun') -> None:
        """Configure a nameserver for the test.

        :param tr: The TestRun to add the nameserver to.
        """
        name = f'dns{self._dns_counter}'
        self._dns = tr.MakeDNServer(name, default='127.0.0.1')
        JA3FingerprintTest._dns_counter += 1

    def _configure_server(self, tr: 'TestRun') -> None:
        """Configure the server to be used in the test.

        :param tr: The TestRun to add the server to.
        """
        name = f'server{self._server_counter}'
        self._server = tr.AddVerifierServerProcess(name, self._replay_file)
        JA3FingerprintTest._server_counter += 1
        self._server.Streams.All += Testers.ContainsExpression("https-request", "Verify the HTTPS request was received.")
        self._server.Streams.All += Testers.ContainsExpression("http2-request", "Verify the HTTP/2 request was received.")

    def _configure_trafficserver(self) -> None:
        """Configure Traffic Server to be used in the test."""
        # Associate ATS with the Test so that metrics can be verified.
        name = f'ts{self._ts_counter}'
        self._ts = Test.MakeATSProcess(name, enable_cache=False, enable_tls=True)
        JA3FingerprintTest._ts_counter += 1
        self._ts.addDefaultSSLFiles()
        self._ts.Disk.ssl_multicert_config.AddLine('dest_ip=* ssl_cert_name=server.pem ssl_key_name=server.key')
        server_port = self._server.Variables.https_port
        self._ts.Disk.remap_config.AddLine(f'map https://https.server.com https://https.backend.com:{server_port}')

        if self._test_remap:
            self._ts.Disk.remap_config.AddLine(
                f'map https://http2.server.com https://http2.backend.com:{server_port} '
                '@plugin=ja3_fingerprint.so @pparam=--ja3log')
        else:
            arguments = '--ja3log --ja3raw'
            if self._modify_incoming:
                arguments += ' --modify-incoming'
            self._ts.Disk.plugin_config.AddLine(f'ja3_fingerprint.so {arguments}')
            self._ts.Disk.remap_config.AddLine(f'map https://http2.server.com https://http2.backend.com:{server_port}')

        self._ts.Disk.records_config.update(
            {
                'proxy.config.ssl.server.cert.path': self._ts.Variables.SSLDir,
                'proxy.config.ssl.server.private_key.path': self._ts.Variables.SSLDir,
                'proxy.config.ssl.client.verify.server.policy': 'PERMISSIVE',
                'proxy.config.dns.nameservers': f"127.0.0.1:{self._dns.Variables.Port}",
                'proxy.config.dns.resolv_conf': 'NULL',
                'proxy.config.diags.debug.enabled': 1,
                'proxy.config.diags.debug.tags': 'http|ja3_fingerprint',
            })

        ja3log_path = os.path.join(self._ts.Variables.LOGDIR, "ja3_fingerprint.log")
        self._ts.Disk.File(ja3log_path, id='ja3_log')

        if self._test_remap:
            # Only the http2 request should be logged because only that request
            # had the plugin configured for that remap rule.
            regex = r'(.*JA3.*MD5){1}'
        else:
            regex = r'(.*JA3.*MD5){2}'

        self._ts.Disk.ja3_log.Content += Testers.ContainsExpression(
            regex, "Verify the JA3 log contains a JA3 line.", reflags=re.MULTILINE | re.DOTALL)

    def _configure_client(self, tr: 'TestRun') -> None:
        """Configure the TestRun.

        :param tr: The TestRun to add the client to.
        """
        name = f'client{self._client_counter}'
        p = tr.AddVerifierClientProcess(
            name,
            self._replay_file,
            http_ports=[self._ts.Variables.port],
            https_ports=[self._ts.Variables.ssl_port],
            other_args='--thread-limit 1')
        JA3FingerprintTest._client_counter += 1

        p.StartBefore(self._dns)
        p.StartBefore(self._server)
        p.StartBefore(self._ts)
        tr.StillRunningAfter = self._ts

        p.Streams.All += Testers.ContainsExpression("https-response", "Verify the HTTPS response was received.")
        p.Streams.All += Testers.ContainsExpression("http2-response", "Verify the HTTP/2 response was received.")

    def _await_ja3log(self) -> None:
        """Await the creation of the JA3 log."""
        tr = Test.AddTestRun('Await the contents of the JA3 log.')

        waiter = tr.Processes.Process('waiter', 'sleep 30')
        ja3_path = self._ts.Disk.ja3_log.AbsPath
        waiter.Ready = When.FileContains(ja3_path, "JA3")

        p = tr.Processes.Default
        p.Command = f'echo await {ja3_path} creation'
        p.StartBefore(waiter)

    def _verify_internal_headers(self) -> None:
        """Verify that the correct headers were modified."""

        # We use the grep command to get the small snippet of output we want
        # from the traffic.out file. Gold file matching against long files, like
        # traffic.out, is exceedingly slow.

        tr = Test.AddTestRun('Verify the internal client request headers.')
        traffic_out = self._ts.Disk.traffic_out.AbsPath
        p = tr.Processes.Default
        p.Command = f'grep --after-context=20 "Incoming Request" {traffic_out}'

        if self._modify_incoming:
            p.Streams.All += "modify-incoming-client.gold"
        else:
            p.Streams.All += "modify-sent-client.gold"

        tr = Test.AddTestRun('Verify the internal proxy request headers.')
        p = tr.Processes.Default
        p.Command = f'grep --after-context=20 "Proxy\'s Request after hooks" {traffic_out}'

        if self._modify_incoming:
            p.Streams.All += "modify-incoming-proxy.gold"
        else:
            p.Streams.All += "modify-sent-proxy.gold"


JA3FingerprintTest(test_remap=False, modify_incoming=False)
JA3FingerprintTest(test_remap=True, modify_incoming=False)
JA3FingerprintTest(test_remap=False, modify_incoming=True)



'''
Implement an OrderedSetQueue
'''
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import collections
try:
    collectionsAbc = collections.abc
except AttributeError:
    collectionsAbc = collections

try:
    import queue as Queue
except ImportError:
    import Queue

#
# This is borrowed from the following (MIT licensed) recipe:
# https://code.activestate.com/recipes/576694/
#


class OrderedSet(collectionsAbc.MutableSet):

    def __init__(self, iterable=None):
        self.end = end = []
        end += [None, end, end]  # sentinel node for doubly linked list
        self.map = {}  # key --> [key, prev, next]
        if iterable is not None:
            self |= iterable

    def __len__(self):
        return len(self.map)

    def __contains__(self, key):
        return key in self.map

    def add(self, key):
        if key not in self.map:
            end = self.end
            curr = end[1]
            curr[2] = end[1] = self.map[key] = [key, curr, end]

    def discard(self, key):
        if key in self.map:
            key, prev, next = self.map.pop(key)
            prev[2] = next
            next[1] = prev

    def __iter__(self):
        end = self.end
        curr = end[2]
        while curr is not end:
            yield curr[0]
            curr = curr[2]

    def __reversed__(self):
        end = self.end
        curr = end[1]
        while curr is not end:
            yield curr[0]
            curr = curr[1]

    # We set last=False so that FIFO is the default behavior.
    def pop(self, last=False):
        if not self:
            raise KeyError('set is empty')
        key = self.end[1][0] if last else self.end[2][0]
        self.discard(key)
        return key

    def __repr__(self):
        if not self:
            return '%s()' % (self.__class__.__name__,)
        return '%s(%r)' % (self.__class__.__name__, list(self))

    def __eq__(self, other):
        if isinstance(other, OrderedSet):
            return len(self) == len(other) and list(self) == list(other)
        return set(self) == set(other)


class OrderedSetQueue(Queue.Queue):
    """
    Make use of the OrderedSet to make a FIFO Queue.Queue that only has unique
    elements in it.
    """

    def _init(self, maxsize):
        self.queue = OrderedSet()

    def _put(self, item):
        self.queue.add(item)

    def _get(self):
        return self.queue.pop()



'''
Test communication to origin with H2
'''
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

Test.Summary = '''
Test communication to origin with H2
'''

Test.ContinueOnFail = True

#
# Communicate to origin with HTTP/2
#
ts = Test.MakeATSProcess("ts", enable_tls="true")

# add ssl materials like key, certificates for the server
ts.addDefaultSSLFiles()
replay_file = "replay_h2origin/"
server = Test.MakeVerifierServerProcess("h2-origin", replay_file)
server_expect = Test.MakeVerifierServerProcess("server-expect", "expect_100_continue.yaml")
ts.Disk.records_config.update(
    {
        'proxy.config.ssl.server.cert.path': '{0}'.format(ts.Variables.SSLDir),
        'proxy.config.ssl.server.private_key.path': '{0}'.format(ts.Variables.SSLDir),
        'proxy.config.diags.debug.enabled': 1,
        'proxy.config.diags.debug.tags': 'http',
        'proxy.config.exec_thread.autoconfig.enabled': 0,
        # Allow for more parallelism
        'proxy.config.exec_thread.limit': 4,
        'proxy.config.ssl.client.alpn_protocols': 'h2,http/1.1',
        # Sticking with thread pool because global pool does not work with h2
        'proxy.config.http.server_session_sharing.pool': 'thread',
        'proxy.config.http.server_session_sharing.match': 'ip,sni,cert',
        'proxy.config.ssl.client.verify.server.policy': 'PERMISSIVE',
    })

ts.Disk.remap_config.AddLines(
    [f'map /expect http://127.0.0.1:{server_expect.Variables.http_port}', f'map / https://127.0.0.1:{server.Variables.https_port}'])
ts.Disk.ssl_multicert_config.AddLine('dest_ip=* ssl_cert_name=server.pem ssl_key_name=server.key')

ts.Disk.logging_yaml.AddLines(
    '''
logging:
  formats:
    - name: testformat
      format: '%<cqtq> %<ttms> %<crc> %<pssc> %<{uuid}cqh> %<cqpv> %<sqpv> %<cqssl> %<cqtr> %<pqssl> %<sstc>'
  logs:
    - mode: ascii
      format: testformat
      filename: squid
'''.split("\n"))

tr = Test.AddTestRun("Test traffic to origin using HTTP/2")
tr.Processes.Default.StartBefore(server)
tr.Processes.Default.StartBefore(ts)
tr.AddVerifierClientProcess("client", replay_file, http_ports=[ts.Variables.port], https_ports=[ts.Variables.ssl_port])
tr.StillRunningAfter = ts
tr.TimeOut = 60

# A regression test for #9857.
tr = Test.AddTestRun("Test an empty body POST request with an Expect: 100-continue header")
tr.AddVerifierClientProcess(
    "client-expect", "expect_100_continue.yaml", http_ports=[ts.Variables.port], https_ports=[ts.Variables.ssl_port])
tr.Processes.Default.StartBefore(server_expect)
tr.Processes.Default.ReturnCode = 0

tr = Test.AddTestRun("Wait for the squid.log to be written")
timeout = 30
watcher = tr.Processes.Process("watcher")
watcher.Command = f"sleep {timeout}"
watcher.Ready = When.FileContains(ts.Disk.squid_log.Name, r'14 http/1.1 http/2')
watcher.TimeOut = timeout
tr.StillRunningAfter = ts
tr.StillRunningAfter = server
tr.TimeOut = timeout
tr.Processes.Default.StartBefore(watcher)
tr.Processes.Default.Command = 'echo await_squid_log'
tr.Processes.Default.ReturnCode = 0

# UUIDs 1-4 should be http/1.1 clients and H2 origin
# UUIDs 5-9 should be http/2 clients and H2 origins
ts.Disk.squid_log.Content = Testers.ContainsExpression(" [1-4] http/1.1 http/2", "cases 1-4 request http/1.1")
ts.Disk.squid_log.Content += Testers.ExcludesExpression(" [1-4] http/2 http/2", "cases 1-4 request http/1.1")
ts.Disk.squid_log.Content += Testers.ContainsExpression(" 1[1-4] http/1.1 http/2", "cases 12-14 request http/1.1")
ts.Disk.squid_log.Content += Testers.ExcludesExpression(" 1[2-4] http/2 http/2", "cases 12-14 request http/1.1")
ts.Disk.squid_log.Content += Testers.ContainsExpression(" [5-9] http/2 http/2", "cases 5-11 request http/2")
ts.Disk.squid_log.Content += Testers.ExcludesExpression(" [5-9] http/1.1 http/2", "cases 5-11 request http/2")
ts.Disk.squid_log.Content += Testers.ContainsExpression(" 1[0-1] http/2 http/2", "cases 5-11 request http/2")
ts.Disk.squid_log.Content += Testers.ExcludesExpression(" 1[0-1] http/1.1 http/2", "cases 5-11 request http/2")

tr = Test.AddTestRun("Test HTTP method Metrics")
tr.Processes.Default.Command = (
    f"{Test.Variables.AtsTestToolsDir}/stdout_wait" + " 'traffic_ctl metric get" + " proxy.process.http.get_requests" +
    " proxy.process.http.post_requests" + " proxy.process.http.put_requests'" +
    f" {Test.TestDirectory}/gold/http-request-method-metrics.gold")
# Need to copy over the environment so traffic_ctl knows where to find the unix
# domain socket
tr.Processes.Default.Env = ts.Env
tr.Processes.Default.ReturnCode = 0
tr.StillRunningAfter = ts



'''
Implement an HTTP/2 server that monitors DATA frame statistics.
'''
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import argparse
import logging
import math
import statistics
import sys
import time
from OpenSSL.SSL import Error as SSLError
from OpenSSL.SSL import SysCallError as SSLSysCallError
import threading

import eventlet
from eventlet.green.OpenSSL import SSL, crypto
from h2.config import H2Configuration
from h2.connection import H2Connection
from h2.events import StreamEnded, RequestReceived, DataReceived, StreamReset, ConnectionTerminated
from h2.errors import ErrorCodes as H2ErrorCodes
from h2.exceptions import StreamClosedError, StreamIDTooLowError

from typing import Dict, List, Optional, Set


def get_body_text() -> bytes:
    """Create a body of text to send in the response."""
    body_items: List[str] = []
    for i in range(100):
        # Create a chunk of 0 padded bytes, followed by a space.
        chunk_payload = f'{i:06x} '.encode("utf-8")
        body_items.append(chunk_payload)
    return b''.join(body_items)


class RequestInfo:
    """POD for request headers, etc."""

    def __init__(self, stream_id: int):
        self.stream_id: int = stream_id
        self.headers: Dict[bytes, bytes] = None
        self.body_bytes: bytes = None


class ResponseInfo:
    """POD for response headers, etc."""

    def __init__(
            self,
            status: int,
            headers: Dict[bytes, bytes],
            body: bytes,
            trailers: Dict[bytes, bytes] = None,
            errors: List[str] = None):
        self.status_code: int = status
        self.headers: Dict[bytes, bytes] = headers
        self.body_bytes: bytes = body
        self.trailers: Dict[bytes, bytes] = trailers
        self.errors: List[str] = errors


class Http2ConnectionManager:
    """Manages a single HTTP/2 connection."""

    def __init__(self, sock: eventlet.greenio.GreenSocket):
        listening_config = H2Configuration(client_side=False, validate_inbound_headers=False)
        self.tls = threading.local()
        self.sock = sock
        self.sock.settimeout(1.0)
        self.listening_conn: H2Connection = H2Connection(config=listening_config)
        self.requests: Dict[int, RequestInfo] = {}
        self.completed_stream_ids: Set[int] = set()

        # Delay times in ms between each data frame.
        self._data_delays: List[int] = []
        # The last time in ms since epoch that a packet was received.
        self.last_packet_time: int = 0

    def _send_responses(self, responses: Dict[int, ResponseInfo]) -> None:
        """Send any responses that have been generated.

        :param responses: A dictionary of responses we wish to send.
        """
        responded_streams = []
        for stream_id, response in responses.items():
            try:
                self.listening_conn.send_headers(stream_id, response.headers)

                send_window = self.listening_conn.local_flow_control_window(stream_id)
                body_size = len(response.body_bytes)
                bytes_to_send = min(send_window, body_size)
                if bytes_to_send < body_size:
                    raise ValueError(
                        f'We do not have a big enough window: body size of {body_size} bytes vs {send_window} byte window')
                # Send one byte at a time, every millisecond.
                bytes_sent = 0
                while bytes_to_send > 0:
                    chunk_size = 1
                    byte_to_send = response.body_bytes[bytes_sent:bytes_sent + chunk_size]
                    logging.debug(f'Sending {byte_to_send}')
                    self.listening_conn.send_data(stream_id, byte_to_send)
                    self.sock.sendall(self.listening_conn.data_to_send())
                    bytes_sent += chunk_size
                    bytes_to_send -= chunk_size
                    time.sleep(0.001)

                self.listening_conn.send_data(stream_id, response.body_bytes, end_stream=False if response.trailers else True)
                if response.trailers is not None:
                    self.listening_conn.send_headers(stream_id, response.trailers, end_stream=True)
                responded_streams.append(stream_id)

            except StreamClosedError as e:
                logging.debug(e)
            except StreamIDTooLowError as e:
                logging.debug(e)
        try:
            # Send the responses we added to the listening_conn.
            self.sock.sendall(self.listening_conn.data_to_send())
        except (SSLError, SSLSysCallError) as e:
            logging.debug(f'Ignoring sock.sendall exception for now: {e}')

        # Clean up any responses we sent.
        for stream_id in responded_streams:
            del responses[stream_id]

    def _receive_data(self, responses: Dict[int, ResponseInfo]) -> Optional[bytes]:
        """Receive data from the socket.

        :param responses: A dictionary of stream IDs to responses that have accumulated.

        :return: The data received, or None if the connection for the socket has closed.
        """
        data: Optional[bytes] = None
        while not data:
            try:
                logging.debug('Receiving data on the socket.')
                data = self.sock.recv(65535)
            except SSLError:
                logging.debug('recv error: the socket is closed.')
                return None
            except TimeoutError:
                # Take time to send any responses we've generated.
                self._send_responses(responses)

                # Loop back around to receive more data.
                logging.debug('Timeout, waiting again for more data.')
                continue
        return data

    def _process_events(self, events: List, responses: Dict[int, ResponseInfo]) -> None:
        """Process events from the H2 connection.

        :param events: The events to process.
        :param responses: A dictionary of stream IDs to responses that have accumulated.
        """
        have_counted_data_delay = False
        for event in events:
            if hasattr(event, 'stream_id'):
                stream_id = event.stream_id
                if stream_id not in self.requests:
                    self.requests[stream_id] = RequestInfo(stream_id)

                request_info = self.requests[stream_id]

                if isinstance(event, DataReceived):
                    if request_info.body_bytes is None:
                        request_info.body_bytes = b''
                    logging.debug(f'Got data for stream {stream_id}: {event.data.decode()}')
                    request_info.body_bytes += event.data

                    if not have_counted_data_delay:
                        ms_since_last_packet = (time.perf_counter_ns() - self.last_packet_time) / (1000 * 1000)
                        logging.debug(f'Counting data delay for stream {stream_id}: {ms_since_last_packet} ms')
                        self._data_delays.append(ms_since_last_packet)
                        self.last_packet_time = time.perf_counter_ns()
                        have_counted_data_delay = True

                if isinstance(event, RequestReceived):
                    logging.info(f'Incoming request received for stream {event.stream_id}.')
                    logging.debug(f'Headers received: {event.headers}')
                    request_info.headers = event.headers
                    self.last_packet_time = time.perf_counter_ns()

                if isinstance(event, StreamReset):
                    self.completed_stream_ids.add(stream_id)
                    err = H2ErrorCodes(event.error_code).name
                    logging.debug(f'Received RST_STREAM frame with error code {err} on stream {event.stream_id}.')
                    if stream_id not in responses.keys():
                        response = self._process_request(request_info)
                        if response is not None:
                            responses[stream_id] = response

                if isinstance(event, StreamEnded):
                    logging.debug('StreamEnded')
                    self.completed_stream_ids.add(stream_id)
                    if stream_id not in responses.keys():
                        response = self._process_request(request_info)
                        if response is not None:
                            responses[stream_id] = response

            else:
                if isinstance(event, ConnectionTerminated):
                    err = H2ErrorCodes(event.error_code).name
                    logging.debug(f'Received GOAWAY frame with error code {err} on with last stream id {event.last_stream_id}.')
                    self.listening_conn.close_connection()

    def _cleanup_closed_stream_ids(self) -> None:
        """Clean up any closed streams."""
        for stream_id in self.completed_stream_ids:
            try:
                if self.listening_conn.streams[stream_id].closed:
                    del self.requests[stream_id]
            except KeyError:
                pass
        try:
            self.completed_stream_ids = set([id for id in self.completed_stream_ids if not self.listening_conn.streams[id].closed])
        except KeyError:
            pass

    def run_forever(self):
        self.listening_conn.initiate_connection()

        try:
            self.sock.sendall(self.listening_conn.data_to_send())
        except (SSLError, SSLSysCallError) as e:
            logging.debug(f'Initial sock.sendall exception: {e}')
            return

        responses: Dict[int, ResponseInfo] = {}
        while True:
            data = self._receive_data(responses)
            if not data:
                # Connection ended.
                break

            logging.debug(f'Giving {len(data)} bytes to the h2 connection')
            events = self.listening_conn.receive_data(data)
            self._process_events(events, responses)
            self._cleanup_closed_stream_ids()
            self._send_responses(responses)

            logging.debug('Sending data on the socket')
            try:
                self.sock.sendall(self.listening_conn.data_to_send())
            except (SSLError, SSLSysCallError) as e:
                logging.debug(f'Ignoring end-loop sock.sendall exception for now: {e}')
                pass

    def get_data_delays(self) -> List[int]:
        """Get the DATA frame timing list.

        :return: The list of DATA frame timings.
        """
        return self._data_delays

    def _process_request(self, request: RequestInfo) -> ResponseInfo:
        """Handle a request from a client.

        :return: A response to send back to the client.
        """
        logging.debug(f'Request received for stream id: {request.stream_id}')
        response_headers = [
            (':status'.encode(), '200'.encode()),
            ('content-type'.encode(), 'text/plain'.encode()),
        ]
        response = ResponseInfo(200, response_headers, get_body_text())

        self._print_transaction(request, response)
        return response

    def _print_transaction(self, request: RequestInfo, response: ResponseInfo) -> None:
        """Print the details of the request and response."""

        description = ''
        description += '\n==== REQUEST HEADERS ====\n'
        for k, v in request.headers:
            if isinstance(k, bytes):
                k, v = (k.decode('ascii'), v.decode('ascii'))
            description += f"{k}: {v}\n"

        if request.body_bytes is not None:
            description += f"\n==== REQUEST BODY ====\n{request.body_bytes.decode()}\n"

        description += "\n==== RESPONSE ====\n"
        description += f"{response.status_code}\n"

        description += "\n==== RESPONSE HEADERS ====\n"
        for k, v in response.headers:
            if isinstance(k, bytes):
                k, v = (k.decode('ascii'), v.decode('ascii'))
            description += f"{k}: {v}\n"

        if response.body_bytes is not None:
            description += f"\n==== RESPONSE BODY ====\n{response.body_bytes.decode()}\n"

        if response.trailers is not None:
            description += "\n==== RESPONSE TRAILERS ====\n"
            for k, v in response.trailers:
                if isinstance(k, bytes):
                    k, v = (k.decode('ascii'), v.decode('ascii'))
                description += f"{k}: {k}\n"
        description += "\n==== END ====\n"
        logging.info(description)


def alpn_callback(conn, protos):
    """The OpenSSL callback for selecting the protocol."""
    if b'h2' in protos:
        return b'h2'

    raise RuntimeError("No acceptable protocol offered!")


def servername_callback(conn):
    """The OpenSSL callback for inspecting the SNI."""
    sni = conn.get_servername()
    conn.set_app_data({'sni': sni})
    logging.info(f"Got SNI from client: {sni}")


def run_server(listen_port, https_pem, ca_pem) -> List[int]:
    """Run the HTTP/2 server.

    :param listen_port: The port to listen on.
    :param https_pem: The path to the certificate key.
    :param ca_pem: The path to the CA certificate.

    :return: The list of DATA frame delays.
    """
    options = (SSL.OP_NO_COMPRESSION | SSL.OP_NO_SSLv2 | SSL.OP_NO_SSLv3 | SSL.OP_NO_TLSv1 | SSL.OP_NO_TLSv1_1)
    context = SSL.Context(SSL.TLSv1_2_METHOD)
    context.set_options(options)
    context.set_verify(SSL.VERIFY_NONE, lambda *args: True)
    context.use_privatekey_file(https_pem)
    context.use_certificate_file(https_pem)
    context.set_alpn_select_callback(alpn_callback)
    context.set_tlsext_servername_callback(servername_callback)
    context.set_cipher_list("RSA+AESGCM".encode())
    context.set_tmp_ecdh(crypto.get_elliptic_curve('prime256v1'))

    listening_socket = eventlet.listen(('0.0.0.0', listen_port))
    listening_socket = SSL.Connection(context, listening_socket)
    logging.info(f"Serving HTTP/2 Proxy on 127.0.0.1:{listen_port} with pem '{https_pem}'")
    pool = eventlet.GreenPool()

    while True:
        try:
            new_connection_socket, _ = listening_socket.accept()
            manager = Http2ConnectionManager(new_connection_socket)
            manager.cert_file = https_pem
            manager.ca_file = ca_pem
            pool.spawn_n(manager.run_forever)
        except KeyboardInterrupt as e:
            logging.debug("Handling KeyboardInterrupt")
            return manager.get_data_delays()
        except SystemExit:
            break


def parse_args() -> argparse.Namespace:
    """Parse command line arguments."""
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument('listen_port', type=int, help='Port to listen on.')
    parser.add_argument('cert_key', type=str, help='Path to the certificate key.')
    parser.add_argument('ca_cert', type=str, help='Path to the CA certificate.')
    parser.add_argument('write_timeout', type=int, help='The timeout between sending frames.')
    parser.add_argument('-v', '--verbose', action='store_true', help='Enable verbose logging.')
    return parser.parse_args()


def main() -> int:
    """Start the HTTP/2 server."""
    args = parse_args()

    log_level = logging.DEBUG if args.verbose else logging.INFO
    logging.basicConfig(level=log_level, format='%(asctime)s - %(levelname)s - %(message)s')

    data_delays = run_server(args.listen_port, args.cert_key, args.ca_cert)
    logging.info(f'Smallest delay: {min(data_delays)} ms')
    logging.info(f'Largest delay: {max(data_delays)} ms')
    average = statistics.mean(data_delays)
    logging.info(f'Average delay over {len(data_delays)} reads: {average} ms')
    isclose = math.isclose(average, args.write_timeout, rel_tol=0.2)
    if isclose:
        logging.info(f'Average delay of {average} is within 20% of the expected delay: {args.write_timeout} ms')
    else:
        logging.info(f'Average delay of {average} is not within 20% of the expected delay: {args.write_timeout} ms')
    return 0 if isclose else 1


if __name__ == '__main__':
    sys.exit(main())



'''
'''
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os

Test.Summary = '''
Basic test for qlog using quiche library.
'''

Test.SkipUnless(Condition.HasATSFeature('TS_HAS_QUICHE'))

Generate_Qlog = True
No_Qlog = False


class quiche_qlog_Test:
    replay_file = "replay/basic1.replay.yaml"
    client_counter: int = 0
    ts_counter: int = 0
    server_counter: int = 0

    def __init__(self, name: str):
        """Initialize the test.
        :param name: The name of the test.
        """
        self._name = name
        self._generate_qlog = False

    def with_qlogs(self):
        self._generate_qlog = True
        return self

    def without_qlogs(self):
        self._generate_qlog = False
        return self

    def _configure_server(self, tr: 'TestRun'):
        """Configure the server.

        :param tr: The TestRun object to associate the server process with.
        """
        server = tr.AddVerifierServerProcess(f"server_{quiche_qlog_Test.server_counter}", self.replay_file)
        quiche_qlog_Test.server_counter += 1
        self._server = server

    def _configure_traffic_server(self, tr: 'TestRun'):
        """Configure Traffic Server.

        :param tr: The TestRun object to associate the ts process with.
        """
        ts = Test.MakeATSProcess(f"ts-{quiche_qlog_Test.ts_counter}", enable_quic=True, enable_tls=True)
        self._ts = ts
        quiche_qlog_Test.ts_counter += 1
        self._ts.addDefaultSSLFiles()
        self._ts.Disk.records_config.update(
            '''
        diags:
          debug:
            enabled: 1
            tags: vv_quic|v_quic
        quic:
          no_activity_timeout_in: 3000
        ''')
        if self._generate_qlog:
            self._ts.Disk.records_config.update(
                '''
                  quic:
                    qlog:
                      file_base: log/test_qlog # we expect to have log/test_qlog-<TRACE ID>.sqlog
                  ''')
        self._ts.Disk.ssl_multicert_config.AddLine(
            f'dest_ip=* ssl_cert_name={ts.Variables.SSLDir}/server.pem ssl_key_name={ts.Variables.SSLDir}/server.key')

        self._ts.Disk.remap_config.AddLine(f'map / http://127.0.0.1:{self._server.Variables.http_port}')

    def test(self):
        """Run the test."""
        tr = Test.AddTestRun(self._name)
        self._configure_server(tr)
        self._configure_traffic_server(tr)

        tr.Processes.Default.StartBefore(self._server)
        tr.Processes.Default.StartBefore(self._ts)
        tr.AddVerifierClientProcess(
            f"client-{quiche_qlog_Test.client_counter}", self.replay_file, http3_ports=[self._ts.Variables.ssl_port])
        quiche_qlog_Test.client_counter += 1
        tr.Processes.Default.ReturnCode = 0

        # If requested we will copy the file to a known name as the runtime name is not know
        # by the test. So we will rename them all and grab the first one for validation.
        test_run = Test.AddTestRun(f"{self._name} : check qlog files")
        qlog_base_name = "test_qlog"
        rename_script = os.path.join(Test.TestDirectory, 'rename_qlog.sh')
        test_run.Processes.Default.Command = f'sleep 5; bash {rename_script} {self._ts.Variables.LOGDIR} {qlog_base_name}'
        test_run.Processes.Default.ReturnCode = 0 if self._generate_qlog else 2  # exit 2 is what we want if no qlog was generated.

        # Basic valdation
        if self._generate_qlog:
            tr = Test.AddTestRun("Check qlog content")
            file = os.path.join(self._ts.Variables.LOGDIR, "1.sqlog")
            f = tr.Disk.File(file)
            tr.Processes.Default.Command = "echo 0"
            tr.Processes.Default.ReturnCode = 0
            f.Content = Testers.IncludesExpression('"title":"Apache Traffic Server"', 'Should include this basic text')


quiche_qlog_Test("Generate qlog test").with_qlogs().test()
quiche_qlog_Test("Do not generate qlog").without_qlogs().test()



#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

Test.Summary = '''
Test command: traffic_ctl config describe proxy.config.http.background_fill_completed_threshold (YTSATS-3309)
'''
Test.testName = 'Float in conf_remap Config Test'

ts = Test.MakeATSProcess("ts")

ts.Disk.MakeConfigFile('conf_remap.yaml').update('''
records:
  http:
    background_fill_completed_threshold: !!float '0.5'
''')

ts.Disk.remap_config.AddLine(
    f"map http://cdn.example.com/ http://origin.example.com/ @plugin=conf_remap.so @pparam={Test.RunDirectory}/ts/config/conf_remap.yaml"
)

tr = Test.AddTestRun("traffic_ctl command")
tr.Env = ts.Env
tr.TimeOut = 5
tr.StillRunningAfter = ts

p = tr.Processes.Default
p.Command = f"traffic_ctl config describe proxy.config.http.background_fill_completed_threshold"
p.ReturnCode = 0
p.StartBefore(Test.Processes.ts)



'''
Test 1 preaccept, 1 sni, and 1 cert callback (with delay)
'''
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os
import re

Test.Summary = '''
Test different combinations of TLS handshake hooks to ensure they are applied consistently.
'''

ts = Test.MakeATSProcess("ts", enable_tls=True)
server = Test.MakeOriginServer("server")
request_header = {"headers": "GET / HTTP/1.1\r\nHost: www.example.com\r\n\r\n", "timestamp": "1469733493.993", "body": ""}
# desired response form the origin server
response_header = {"headers": "HTTP/1.1 200 OK\r\nConnection: close\r\n\r\n", "timestamp": "1469733493.993", "body": ""}
server.addResponse("sessionlog.json", request_header, response_header)

ts.addDefaultSSLFiles()

ts.Disk.records_config.update(
    {
        'proxy.config.diags.debug.enabled': 1,
        'proxy.config.diags.show_location': 0,
        'proxy.config.diags.debug.tags': 'ssl_hook_test',
        'proxy.config.ssl.server.cert.path': '{0}'.format(ts.Variables.SSLDir),
        'proxy.config.ssl.server.private_key.path': '{0}'.format(ts.Variables.SSLDir),
    })

ts.Disk.ssl_multicert_config.AddLine('dest_ip=* ssl_cert_name=server.pem ssl_key_name=server.key')

ts.Disk.remap_config.AddLine(
    'map https://example.com:{0} http://127.0.0.1:{1}'.format(ts.Variables.ssl_port, server.Variables.Port))

Test.PrepareTestPlugin(os.path.join(Test.Variables.AtsTestPluginsDir, 'ssl_hook_test.so'), ts, '-cert=1 -sni=1 -preaccept=1')

tr = Test.AddTestRun("Test one sni, one preaccept, and one cert hook")
tr.Processes.Default.StartBefore(server)
tr.Processes.Default.StartBefore(Test.Processes.ts)
tr.StillRunningAfter = ts
tr.StillRunningAfter = server
tr.Processes.Default.Command = 'curl -k -H \'host:example.com:{0}\' https://127.0.0.1:{0}'.format(ts.Variables.ssl_port)
tr.Processes.Default.ReturnCode = 0
tr.Processes.Default.Streams.stdout = "gold/preaccept-1.gold"

ts.Disk.traffic_out.Content = "gold/ts-preaccept1-sni1-cert1.gold"
snistring = "SNI callback 0"
preacceptstring = "Pre accept callback 0"
certstring = "Cert callback 0"
ts.Disk.traffic_out.Content = Testers.ContainsExpression(
    r"\A(?:(?!{0}).)*{0}(?!.*{0}).*\Z".format(snistring), "SNI message appears only once", reflags=re.S | re.M)
# the preaccept may get triggered twice because the test framework creates a TCP connection before handing off to traffic_server
ts.Disk.traffic_out.Content += Testers.ContainsExpression(
    r"\A(?:(?!{0}).)*{0}.*({0})?(?!.*{0}).*\Z".format(preacceptstring),
    "Pre accept message appears only once or twice",
    reflags=re.S | re.M)
ts.Disk.traffic_out.Content += Testers.ContainsExpression(
    r"\A(?:(?!{0}).)*{0}(?!.*{0}).*\Z".format(certstring), "Cert message appears only once", reflags=re.S | re.M)

tr.Processes.Default.TimeOut = 15
tr.TimeOut = 15



# @file
#
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
#  Copyright 2020, Verizon Media
#
'''
Basic smoke tests.
'''
Test.Summary = '''
Test basic functions and directives via remaps.
'''

Test.SkipUnless(Condition.PluginExists("txn_box.so"))

replay_file = 'smoke-2.replay.yaml'
Test.TxnBoxTestAndRun(
    "Smoke 2 Test",
    replay_file,
    config_path='Auto',
    config_key="meta.txn_box.global",
    remap=[
        ['http://alpha.ex/', ('--key=meta.txn_box.remap.alpha', replay_file)],
        ['http://bravo.ex/', ('--key=meta.txn_box.remap.bravo', replay_file)]
    ])



'''
'''
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os
import ports

Test.Summary = '''
Ensure that handshake fails if invalid alpn string is offered
'''

# Only later versions of openssl support the `-alpn` option.
Test.SkipUnless(Condition.HasOpenSSLVersion('1.1.1'))

# Define default ATS
ts = Test.MakeATSProcess("ts", enable_tls=True)

# add ssl materials like key, certificates for the server
ts.addSSLfile("ssl/server.pem")
ts.addSSLfile("ssl/server.key")

# Make sure the TS server certs are different from the origin certs
ts.Disk.ssl_multicert_config.AddLine('dest_ip=* ssl_cert_name=server.pem ssl_key_name=server.key')

# Case 1, global config policy=permissive properties=signature
#         override for foo.com policy=enforced properties=all
ts.Disk.records_config.update(
    {
        'proxy.config.ssl.server.cert.path': '{0}'.format(ts.Variables.SSLDir),
        'proxy.config.ssl.server.private_key.path': '{0}'.format(ts.Variables.SSLDir),
    })

tr = Test.AddTestRun("alpn banana")
tr.Processes.Default.Command = "openssl s_client -ign_eof -alpn=banana -connect 127.0.0.1:{}".format(ts.Variables.ssl_port)
tr.ReturnCode = 1  # Should fail
tr.Processes.Default.StartBefore(Test.Processes.ts)
tr.StillRunningAfter = ts
tr.Processes.Default.Streams.All += Testers.IncludesExpression("No ALPN negotiated", "Banana should not match a negotiation")

tr = Test.AddTestRun("alpn http/1.1")
tr.Processes.Default.Command = "printf 'GET / HTTP/1.1\r\n\r\n' | openssl s_client -ign_eof -alpn=http/1.1 -connect 127.0.0.1:{}".format(
    ts.Variables.ssl_port)
tr.ReturnCode = 0
tr.StillRunningAfter = ts
tr.Processes.Default.Streams.All += Testers.IncludesExpression("ALPN protocol: http/1.1", "Successful ALPN")
tr.Processes.Default.Streams.All += Testers.IncludesExpression("HTTP/1.1 400 Host Header Required", "Processed the request")

tr = Test.AddTestRun("no alpn")
tr.Processes.Default.Command = "printf 'GET / HTTP/1.1\r\n\r\n' | openssl s_client -ign_eof -connect 127.0.0.1:{}".format(
    ts.Variables.ssl_port)
tr.ReturnCode = 0
tr.StillRunningAfter = ts
tr.Processes.Default.Streams.All += Testers.IncludesExpression("No ALPN negotiated", "No ALPN offered, none negotiated")
tr.Processes.Default.Streams.All += Testers.IncludesExpression("HTTP/1.1 400 Host Header Required", "Processed the request")

tr = Test.AddTestRun("alpn h2")
tr.Processes.Default.Command = "curl -k --http2 -v -o /dev/null https://127.0.0.1:{}".format(ts.Variables.ssl_port)
tr.ReturnCode = 0
tr.StillRunningAfter = ts
tr.Processes.Default.Streams.All += Testers.IncludesExpression("ALPN. server accepted.*h2", "negotiated h2")
tr.Processes.Default.Streams.All += Testers.IncludesExpression("HTTP/2 404 ", "Good response")



# @file
#
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
#  Copyright 2020, Verizon Media
#
import os.path

Test.Summary = '''
Multi-bucketing (style 3).
'''

Test.SkipUnless(Condition.PluginExists("txn_box.so"))

RepeatCount = 1000
CFG_PATH = "multi-ramp-3.cfg.yaml"
tr = Test.TxnBoxTestAndRun(
    "Multi bucketing 3",
    "multi-ramp.replay.yaml",
    remap=[['http://base.ex/', 'http://base.ex/', [CFG_PATH]], ['https://base.ex/', 'https://base.ex/', [CFG_PATH]]],
    verifier_client_args='--verbose info --format "{{url}}" --repeat {}'.format(RepeatCount),
    verifier_server_args='--verbose info --format "{url}"',
    enable_tls=True)

with open(f"{tr.TestDirectory}/multi_ramp_common.py") as f:
    code = compile(f.read(), "multi_ramp_common.py", 'exec')
    exec(code)

ramp_test_fixup(tr)
ts = tr.Variables.TXNBOX_TS
ts.Setup.Copy(CFG_PATH, ts.Variables.CONFIGDIR)



'''
Test cache lookup results in CACHE condition
'''
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
Test.ContinueOnFail = True
ts = Test.MakeATSProcess("ts")
server = Test.MakeOriginServer("server")

Test.testName = "CACHE"

# Request from client
request_header = {"headers": "GET / HTTP/1.1\r\nHost: www.example.com\r\n\r\n", "timestamp": "1469733493.993", "body": ""}
# Expected response from the origin server
response_header = {
    "headers": "HTTP/1.1 200 OK\r\nConnection: close\r\nCache-Control: max-age=10,public\r\n\r\n",
    "timestamp": "1469733493.993",
    "body": "CACHED"
}

# add request/response
server.addResponse("sessionlog.log", request_header, response_header)
ts.Disk.records_config.update({
    'proxy.config.diags.debug.enabled': 1,
    'proxy.config.diags.debug.tags': 'http',
})

# This rule adds the cache header with results of the cache lookup results
# (ie. "hit-fresh", "hit-stale", "miss", "none")
ts.Setup.CopyAs('rules/rule_add_cache_result_header.conf', Test.RunDirectory)

ts.Disk.plugin_config.AddLine('header_rewrite.so {0}/rule_add_cache_result_header.conf'.format(Test.RunDirectory))
ts.Disk.remap_config.AddLine('map http://www.example.com http://127.0.0.1:{0}'.format(server.Variables.Port))
ts.Disk.remap_config.AddLine('map / http://127.0.0.1:{0}'.format(server.Variables.Port))

# Commands to get the following response headers
# 1. miss (empty cache)
# 2. hit-fresh (content served within 10s cache)
# 3. hit-stale (waited 15s, after 10s cache)
# 2. hit-fresh (content served within 10s cache)
curlRequest = (
    'curl -s -v -H "Host: www.example.com" http://127.0.0.1:{0};'
    'curl -v -H "Host: www.example.com" http://127.0.0.1:{0};'
    'sleep 15; curl -s -v -H "Host: www.example.com" http://127.0.0.1:{0};'
    'curl -s -v -H "Host: www.example.com" http://127.0.0.1:{0}')

# Test Case
tr = Test.AddTestRun()
tr.Processes.Default.Command = curlRequest.format(ts.Variables.port)
tr.Processes.Default.ReturnCode = 0
tr.Processes.Default.StartBefore(server)
tr.Processes.Default.StartBefore(ts)
tr.Processes.Default.Streams.stderr = "gold/header_rewrite_cond_cache.gold"
tr.StillRunningAfter = server



'''
'''
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

Test.Summary = '''
Test for server_push_preload plugin
'''

Test.SkipUnless(
    Condition.PluginExists('server_push_preload.so'),
    Condition.HasProgram("nghttp", "Nghttp need to be installed on system for this test to work"),
)
Test.testName = "server_push_preload"
Test.ContinueOnFail = True

# ----
# Setup Origin Server
# ----
microserver = Test.MakeOriginServer("microserver")

# index.html
microserver.addResponse(
    "sessionfile.log", {
        "headers": "GET /index.html HTTP/1.1\r\nHost: www.example.com\r\n\r\n",
        "body": ""
    }, {
        "headers":
            "HTTP/1.1 200 OK\r\nConnection: close\r\nLink: </app/style.css>; rel=preload; as=style; nopush\r\nLink: </app/script.js>; rel=preload; as=script\r\n\r\n",
        "body":
            "<html>\r\n<head>\r\n<link rel='stylesheet' type='text/css' href='/app/style.css' />\r\n<script src='/app/script.js'></script>\r\n</head>\r\n<body>\r\nServer Push Preload Test\r\n</body>\r\n</html>\r\n"
    })

# /app/style.css
microserver.addResponse(
    "sessionfile.log", {
        "headers": "GET /app/style.css HTTP/1.1\r\nHost: www.example.com\r\n\r\n",
        "body": ""
    }, {
        "headers": "HTTP/1.1 200 OK\r\nConnection: close\r\n\r\n",
        "body": "body { font-weight: bold; }\r\n"
    })

# /app/script.js
microserver.addResponse(
    "sessionfile.log", {
        "headers": "GET /app/script.js HTTP/1.1\r\nHost: www.example.com\r\n\r\n",
        "body": ""
    }, {
        "headers": "HTTP/1.1 200 OK\r\nConnection: close\r\n\r\n",
        "body": "function do_nothing() { return; }\r\n"
    })

# ----
# Setup ATS
# ----
ts = Test.MakeATSProcess("ts", enable_tls=True)

ts.addDefaultSSLFiles()

ts.Disk.ssl_multicert_config.AddLine('dest_ip=* ssl_cert_name=server.pem ssl_key_name=server.key')

ts.Disk.remap_config.AddLine('map / http://127.0.0.1:{0}/ @plugin=server_push_preload.so'.format(microserver.Variables.Port))

ts.Disk.records_config.update(
    {
        'proxy.config.diags.debug.enabled': 1,
        'proxy.config.diags.debug.tags': 'http2|server_push_preload',
        'proxy.config.ssl.server.cert.path': '{0}'.format(ts.Variables.SSLDir),
        'proxy.config.ssl.server.private_key.path': '{0}'.format(ts.Variables.SSLDir),
        'proxy.config.http2.active_timeout_in': 3,
    })

# ----
# Test Cases
# ----

# Test Case 0: Server Push by Link header
tr = Test.AddTestRun()
tr.Processes.Default.Command = "nghttp -vs --no-dep 'https://127.0.0.1:{0}/index.html'".format(ts.Variables.ssl_port)
tr.Processes.Default.ReturnCode = 0
tr.Processes.Default.StartBefore(microserver, ready=When.PortOpen(microserver.Variables.Port))
tr.Processes.Default.StartBefore(Test.Processes.ts)
tr.Processes.Default.Streams.stdout = "gold/server_push_preload_0_stdout.gold"
tr.StillRunningAfter = microserver



'''
'''
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os

Test.Summary = '''
Test log fields.
'''

ts = Test.MakeATSProcess("ts", enable_cache=False)
server = Test.MakeOriginServer("server")

request_header = {'timestamp': 100, "headers": "GET /test-1 HTTP/1.1\r\nHost: test-1\r\n\r\n", "body": ""}
response_header = {
    'timestamp': 100,
    "headers":
        "HTTP/1.1 200 OK\r\nTest: 1\r\nContent-Type: application/json\r\nConnection: close\r\nContent-Type: application/json\r\n\r\n",
    "body": "Test 1"
}
server.addResponse("sessionlog.json", request_header, response_header)
server.addResponse(
    "sessionlog.json", {
        'timestamp': 101,
        "headers": "GET /test-2 HTTP/1.1\r\nHost: test-2\r\n\r\n",
        "body": ""
    }, {
        'timestamp': 101,
        "headers":
            "HTTP/1.1 200 OK\r\nTest: 2\r\nContent-Type: application/jason\r\nConnection: close\r\nContent-Type: application/json\r\n\r\n",
        "body": "Test 2"
    })
server.addResponse(
    "sessionlog.json", {
        'timestamp': 102,
        "headers": "GET /test-3 HTTP/1.1\r\nHost: test-3\r\n\r\n",
        "body": ""
    }, {
        'timestamp': 102,
        "headers": "HTTP/1.1 200 OK\r\nTest: 3\r\nConnection: close\r\nContent-Type: application/json\r\n\r\n",
        "body": "Test 3"
    })

nameserver = Test.MakeDNServer("dns", default='127.0.0.1')

ts.Disk.records_config.update(
    {
        'proxy.config.net.connections_throttle': 100,
        'proxy.config.dns.nameservers': f"127.0.0.1:{nameserver.Variables.Port}",
        'proxy.config.dns.resolv_conf': 'NULL'
    })
# setup some config file for this server
ts.Disk.remap_config.AddLine('map / http://localhost:{}/'.format(server.Variables.Port))

ts.Disk.logging_yaml.AddLines(
    '''
logging:
  formats:
    - name: custom
      escape: json
      format: '{"foo":"%<{Foo}cqh>","foo-slice":"%<{Foo}cqh[2:-3]>"}'
  logs:
    - filename: field-json-test
      format: custom
'''.split("\n"))

# #########################################################################
# at the end of the different test run a custom log file should exist
# Because of this we expect the testruns to pass the real test is if the
# customlog file exists and passes the format check
Test.Disk.File(os.path.join(ts.Variables.LOGDIR, 'field-json-test.log'), exists=True, content='gold/field-json-test.gold')

# first test is a miss for default
tr = Test.AddTestRun()
# Wait for the micro server
tr.Processes.Default.StartBefore(server)
tr.Processes.Default.StartBefore(nameserver)
# Delay on readiness of our ssl ports
tr.Processes.Default.StartBefore(Test.Processes.ts)

tr.Processes.Default.Command = 'curl --verbose --header "Host: test-1" --header "Foo: ab\td/ef" http://localhost:{0}/test-1'.format(
    ts.Variables.port)
tr.Processes.Default.ReturnCode = 0

tr = Test.AddTestRun()
tr.Processes.Default.Command = 'curl --verbose --header "Host: test-2" --header "Foo: ab\x1fd/ef" http://localhost:{0}/test-2'.format(
    ts.Variables.port)
tr.Processes.Default.ReturnCode = 0

tr = Test.AddTestRun()
tr.Processes.Default.Command = 'curl --verbose --header "Host: test-3" --header "Foo: abc\x7fde" http://localhost:{0}/test-3'.format(
    ts.Variables.port)
tr.Processes.Default.ReturnCode = 0

tr = Test.AddTestRun()
tr.Processes.Default.Command = 'curl --verbose --header "Host: test-2" --header "Foo: ab\x80d/ef" http://localhost:{0}/test-4'.format(
    ts.Variables.port)
tr.Processes.Default.ReturnCode = 0

# Wait for log file to appear, then wait one extra second to make sure TS is done writing it.
test_run = Test.AddTestRun()
test_run.Processes.Default.Command = (
    os.path.join(Test.Variables.AtsTestToolsDir, 'condwait') + ' 60 1 -f ' +
    os.path.join(ts.Variables.LOGDIR, 'field-json-test.log'))
test_run.Processes.Default.ReturnCode = 0



'''
'''
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os

Test.Summary = 'Test TSContScheduleOnPool API'
Test.ContinueOnFail = True

# Define default ATS
ts = Test.MakeATSProcess('ts')

Test.testName = 'Test TSContScheduleOnPool API'

ts.Disk.records_config.update(
    {
        'proxy.config.exec_thread.autoconfig.enabled': 0,
        'proxy.config.exec_thread.autoconfig.scale': 1.5,
        'proxy.config.exec_thread.limit': 32,
        'proxy.config.accept_threads': 1,
        'proxy.config.task_threads': 2,
        'proxy.config.diags.debug.enabled': 1,
        'proxy.config.diags.debug.tags': 'TSContSchedule_test'
    })

# Load plugin
Test.PrepareTestPlugin(os.path.join(Test.Variables.AtsTestPluginsDir, 'cont_schedule.so'), ts, 'pool')

# www.example.com Host
tr = Test.AddTestRun()
tr.Processes.Default.Command = 'printf "Test TSContScheduleOnPool API"'
tr.Processes.Default.ReturnCode = 0
tr.Processes.Default.StartBefore(ts)
ts.Disk.traffic_out.Content = "gold/schedule_on_pool.gold"
ts.Disk.traffic_out.Content += Testers.ExcludesExpression('fail', 'should not contain "fail"')



'''
'''
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import re

Test.Summary = '''
Test tls tickets
'''

# Define default ATS
ts = Test.MakeATSProcess("ts", enable_tls=True)
ts2 = Test.MakeATSProcess("ts2", enable_tls=True)
server = Test.MakeOriginServer("server")

# Add info the origin server responses
request_header = {"headers": "GET / HTTP/1.1\r\nHost: www.example.com\r\n\r\n", "timestamp": "1469733493.993", "body": ""}
response_header = {"headers": "HTTP/1.1 200 OK\r\nConnection: close\r\n\r\n", "timestamp": "1469733493.993", "body": ""}
server.addResponse("sessionlog.json", request_header, response_header)

# add ssl materials like key, certificates for the server
ts.addSSLfile("ssl/server.pem")
ts.addSSLfile("ssl/server.key")
ts2.addSSLfile("ssl/server.pem")
ts2.addSSLfile("ssl/server.key")

ts.Disk.remap_config.AddLine('map / http://127.0.0.1:{0}'.format(server.Variables.Port))
ts2.Disk.remap_config.AddLine('map / http://127.0.0.1:{0}'.format(server.Variables.Port))

ts.Disk.ssl_multicert_config.AddLine('dest_ip=* ssl_cert_name=server.pem ssl_key_name=server.key')
ts2.Disk.ssl_multicert_config.AddLine('dest_ip=* ssl_cert_name=server.pem ssl_key_name=server.key')

ts.Disk.records_config.update(
    {
        'proxy.config.ssl.server.cert.path': '{0}'.format(ts.Variables.SSLDir),
        'proxy.config.ssl.server.private_key.path': '{0}'.format(ts.Variables.SSLDir),
        'proxy.config.exec_thread.autoconfig.scale': 1.0,
        'proxy.config.ssl.server.session_ticket.enable': '1',
        'proxy.config.ssl.server.ticket_key.filename': '../../file.ticket'
    })
ts2.Disk.records_config.update(
    {
        'proxy.config.ssl.server.cert.path': '{0}'.format(ts2.Variables.SSLDir),
        'proxy.config.ssl.server.private_key.path': '{0}'.format(ts2.Variables.SSLDir),
        'proxy.config.ssl.server.session_ticket.enable': '1',
        'proxy.config.exec_thread.autoconfig.scale': 1.0,
        'proxy.config.ssl.server.ticket_key.filename': '../../file.ticket'
    })

tr = Test.AddTestRun("Create ticket")
tr.Setup.Copy('file.ticket')
tr.Command = 'echo -e "GET / HTTP/1.0\r\n" | openssl s_client -tls1_2 -connect 127.0.0.1:{0} -sess_out ticket.out'.format(
    ts.Variables.ssl_port)
tr.ReturnCode = 0
tr.Processes.Default.StartBefore(server)
tr.Processes.Default.StartBefore(Test.Processes.ts)
path1 = tr.Processes.Default.Streams.stdout.AbsPath
tr.StillRunningAfter = server

# Pull out session created in tr to test for session id in tr2


def checkSession(ev):
    retval = False
    f1 = open(path1, 'r')
    f2 = open(path2, 'r')
    err = "Session ids match"
    if not f1 or not f2:
        err = "Failed to open {0} or {1}".format(path1, path2)
        return (retval, "Check that session ids match", err)

    f1Content = f1.read()
    f2Content = f2.read()
    match1 = re.findall('Session-ID: ([0-9A-F]+)', f1Content)
    match2 = re.findall('Session-ID: ([0-9A-F]+)', f2Content)

    if match1 and match2:
        if match1[0] == match2[0]:
            err = "{0} and {1} do match".format(match1[0], match2[0])
            retval = True
        else:
            err = "{0} and {1} do not match".format(match1[0], match2[0])
    else:
        err = "Didn't find session id"
    return (retval, "Check that session ids match", err)


tr2 = Test.AddTestRun("Test ticket")
tr2.Setup.Copy('file.ticket')
tr2.Command = 'echo -e "GET / HTTP/1.0\r\n" | openssl s_client -tls1_2 -connect 127.0.0.1:{0} -sess_in ticket.out'.format(
    ts2.Variables.ssl_port)
tr2.Processes.Default.StartBefore(Test.Processes.ts2)
tr2.ReturnCode = 0
path2 = tr2.Processes.Default.Streams.stdout.AbsPath
tr2.Processes.Default.Streams.All.Content = Testers.Lambda(checkSession)



#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os

Test.Summary = '''
Test traffic_ctl config reload with remap.config .include directive
'''

Test.ContinueOnFail = False

Test.Setup.Copy("wait_reload.sh")

# Define ATS and configure
ts = Test.MakeATSProcess("ts", enable_cache=False)
nameserver = Test.MakeDNServer("dns", default='127.0.0.1')

ts.Disk.File(ts.Variables.CONFIGDIR + "/test.inc", id="test_cfg", typename="ats:config")
ts.Disk.test_cfg.AddLine(
    "map http://example.two/ http://yada.com/ " + "@plugin=conf_remap.so @pparam=proxy.config.url_remap.pristine_host_hdr=1")

ts.Disk.remap_config.AddLine("map http://example.one/ http://yada.com/")
ts.Disk.remap_config.AddLine(".include test.inc")
ts.Disk.remap_config.AddLine("map http://example.three/ http://yada.com/")

# minimal configuration
ts.Disk.records_config.update(
    {
        'proxy.config.diags.debug.enabled': 1,
        'proxy.config.diags.debug.tags': 'regex_remap|url_rewrite|plugin_factory',
        'proxy.config.dns.nameservers': f"127.0.0.1:{nameserver.Variables.Port}",
    })

tr = Test.AddTestRun("Start TS, then update test.inc")
tr.Processes.Default.StartBefore(Test.Processes.ts)
tr.Processes.Default.StartBefore(nameserver)
test_inc_path = ts.Variables.CONFIGDIR + "/test.inc"
tr.Processes.Default.Command = (
    f"rm -f {test_inc_path} ; " + f"echo 'map http://example.four/ http://localhost/ @plugin=generator.so' > {test_inc_path}")
tr.Processes.Default.ReturnCode = 0
tr.StillRunningAfter = ts

tr = Test.AddTestRun("Reload config")
tr.StillRunningAfter = ts
tr.Processes.Default.Command = f'traffic_ctl config reload'
# Need to copy over the environment so traffic_ctl knows where to find the unix domain socket
tr.Processes.Default.Env = ts.Env
tr.Processes.Default.ReturnCode = 0

tr = Test.AddTestRun("Wait for config reload")
tr.Processes.Default.Command = './wait_reload.sh ' + os.path.join(ts.Variables.LOGDIR, 'diags.log')
tr.Processes.Default.ReturnCode = 0
tr.StillRunningAfter = ts

tr = Test.AddTestRun("Get response from generator")
tr.Processes.Default.Command = f'curl --proxy 127.0.0.1:{ts.Variables.port} http://example.four/nocache/5'
tr.Processes.Default.ReturnCode = 0
tr.StillRunningAfter = ts
tr.Processes.Default.Streams.All = Testers.ContainsExpression("xxxxx", "Contains generated text")
tr.Processes.Default.Streams.All += Testers.ExcludesExpression("xxxxxx", "Not too much data")



#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os

Test.Summary = '''
Check whether the HTTP/2 grace shutdown works as intended.
'''

Test.SkipUnless(Condition.HasProxyVerifierVersion('2.8.0'))

pv_server = Test.MakeVerifierServerProcess("pv_server", "http2_close_connection.yaml")

ts = Test.MakeATSProcess('ts', select_ports=True, enable_tls=True)

ts.addDefaultSSLFiles()
ts.Disk.ssl_multicert_config.AddLine("dest_ip=* ssl_cert_name=server.pem ssl_key_name=server.key")
ts.Disk.records_config.update(
    {
        "proxy.config.http.server_ports": f"{ts.Variables.port} {ts.Variables.ssl_port}:ssl",
        'proxy.config.ssl.server.cert.path': f'{ts.Variables.SSLDir}',
        'proxy.config.ssl.server.private_key.path': f'{ts.Variables.SSLDir}',
        'proxy.config.ssl.client.verify.server.policy': 'PERMISSIVE',
        'proxy.config.diags.debug.enabled': 3,
        'proxy.config.diags.debug.tags': 'http',
        'proxy.config.exec_thread.autoconfig.enabled': 0,
        'proxy.config.exec_thread.limit': 4,
        'proxy.config.ssl.client.alpn_protocols': 'h2,http/1.1',
        'proxy.config.http.server_session_sharing.pool': 'thread',
        'proxy.config.http.server_session_sharing.match': 'ip,sni,cert',
    })

ts.Disk.remap_config.AddLines([f'map / https://127.0.0.1:{pv_server.Variables.https_port}'])

Test.PrepareTestPlugin(os.path.join(Test.Variables.AtsTestPluginsDir, 'http2_close_connection.so'), ts)

tr = Test.AddTestRun()
tr.Processes.Default.StartBefore(pv_server)
tr.Processes.Default.StartBefore(ts)
tr.AddVerifierClientProcess(
    "pv_client",
    "http2_close_connection.yaml",
    http_ports=[ts.Variables.port],
    https_ports=[ts.Variables.ssl_port],
    other_args='--thread-limit 1')
tr.Processes.Default.ReturnCode = 0

tr.Processes.Default.Streams.All += Testers.ContainsExpression(
    'Equals Success: Key: "1", Content Data: "body", Value: "server_test_1"', 'Response check')
tr.Processes.Default.Streams.All += Testers.ContainsExpression(
    'Received GOAWAY frame with last stream id 2147483647, error code 0', 'initial GOAWAY frame with last stream id set to max')
tr.Processes.Default.Streams.All += Testers.ContainsExpression(
    'Received GOAWAY frame with last stream id 1, error code 0', 'updated GOAWAY frame with last stream id set to 1')

pv_server.Streams.All += Testers.ExcludesExpression('server_test_2', 'Only one response should be sent')



'''
'''
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os

Test.Summary = '''

'''
Test.SkipUnless(Condition.PluginExists('cookie_remap.so'))
Test.ContinueOnFail = True
Test.testName = "cookie_remap: test connector"

# Define default ATS
ts = Test.MakeATSProcess("ts")

# First server is run during first test and
# second server is run during second test
server = Test.MakeOriginServer("server", ip='127.0.0.10')

request_header = {
    "headers": "GET /cookiematches HTTP/1.1\r\nHost: www.example.com\r\n\r\n",
    "timestamp": "1469733493.993",
    "body": ""
}
# expected response from the origin server
response_header = {"headers": "HTTP/1.1 200 OK\r\nConnection: close\r\n\r\n", "timestamp": "1469733493.993", "body": ""}

# add response to the server dictionary
server.addResponse("sessionfile.log", request_header, response_header)

server2 = Test.MakeOriginServer("server2", ip='127.0.0.11')
request_header2 = {
    "headers": "GET /cookiedoesntmatch HTTP/1.1\r\nHost: www.example.com\r\n\r\n",
    "timestamp": "1469733493.993",
    "body": ""
}
# expected response from the origin server
response_header2 = {"headers": "HTTP/1.1 200 OK\r\nConnection: close\r\n\r\n", "timestamp": "1469733493.993", "body": ""}

# add response to the server dictionary
server2.addResponse("sessionfile.log", request_header2, response_header2)

# Setup the remap configuration
config_path = os.path.join(Test.TestDirectory, "configs/subcookie.txt")
with open(config_path, 'r') as config_file:
    config1 = config_file.read()

ts.Disk.records_config.update(
    {
        'proxy.config.diags.debug.enabled': 1,
        'proxy.config.diags.debug.tags': 'cookie_remap.*|http.*|dns.*',
    })

config1 = config1.replace("$PORT", str(server.Variables.Port))
config1 = config1.replace("$ALTPORT", str(server2.Variables.Port))

ts.Disk.File(ts.Variables.CONFIGDIR + "/subcookie.txt", id="config1")
ts.Disk.config1.WriteOn(config1)

ts.Disk.remap_config.AddLine(
    'map http://www.example.com http://shouldnothit.com/magic @plugin=cookie_remap.so @pparam=config/subcookie.txt')

# Positive test case that remaps because all connected operations pass
tr = Test.AddTestRun("cookie value matches")
# Unlike in other places I am using a single line string because the & seems to
# be interpreted by the autest framework or the shell (tried escaping with \)
tr.Processes.Default.Command = 'curl --proxy 127.0.0.1:{0} "http://www.example.com" -H"Cookie: fpbeta=a=1&b=2&c=3" -H "Proxy-Connection: keep-alive" --verbose '.format(
    ts.Variables.port)
# tr.Processes.Default.Command = '''
# curl
# --proxy 127.0.0.1:{0}
# "http://www.example.com"
# -H\"Cookie: fpbeta=a=1&b=2&c=3\"
# -H "Proxy-Connection: keep-alive"
# --verbose
# '''.format(ts.Variables.port)
tr.Processes.Default.ReturnCode = 0
tr.Processes.Default.StartBefore(server, ready=When.PortOpen(server.Variables.Port))
tr.Processes.Default.StartBefore(Test.Processes.ts)
tr.StillRunningAfter = ts

server.Streams.All = "gold/matchcookie.gold"

# Negative test case that doesn't remap because not all subops pass
tr = Test.AddTestRun("cookie value doesn't match")
tr.Processes.Default.Command = 'curl --proxy 127.0.0.1:{0} "http://www.example.com" -H"Cookie: fpbeta=a=1&b=2&c=4" -H "Proxy-Connection: keep-alive" --verbose '.format(
    ts.Variables.port)
tr.Processes.Default.ReturnCode = 0
tr.Processes.Default.StartBefore(server2, ready=When.PortOpen(server2.Variables.Port))
tr.StillRunningAfter = ts

server2.Streams.All = "gold/wontmatchcookie.gold"
